from __future__ import annotations

import functools
import logging
import os
import re
from datetime import datetime
from typing import TYPE_CHECKING, Any, BinaryIO, Dict, Literal

from gql import Client as GQLClient
from gql.transport.requests import RequestsHTTPTransport

from vectice.__version__ import __version__
from vectice.api._auth import Auth
from vectice.api._utils import get_asset_type, read_nodejs_date
from vectice.api.attachment import AttachmentApi
from vectice.api.columns_description import ColumnsDescriptionApi
from vectice.api.compatibility import CompatibilityApi
from vectice.api.gql_attachment import GqlAttachmentApi
from vectice.api.gql_dataset import GqlDatasetApi
from vectice.api.gql_feature_flag import GqlFeatureFlagApi
from vectice.api.gql_issue import GqlIssueApi
from vectice.api.gql_lineage import GqlLineageApi
from vectice.api.gql_metric import GqlMetricApi
from vectice.api.gql_model import GqlModelApi
from vectice.api.gql_organization import GqlOrganizationApi
from vectice.api.gql_phase import GqlPhaseApi
from vectice.api.gql_property import GqlPropertyApi
from vectice.api.gql_report import GqlReportApi
from vectice.api.gql_review import GqlReviewApi
from vectice.api.gql_user_workspace_api import UserAndDefaultWorkspaceApi
from vectice.api.http_error_handlers import MissingReferenceError, VecticeException
from vectice.api.iteration import IterationApi
from vectice.api.json import (
    ArtifactName,
    ModelRegisterInput,
    ModelRegisterOutput,
    ModelVersionOutput,
    ModelVersionStatus,
    PagedResponse,
    PropertyInput,
    RequirementOutput,
)
from vectice.api.json.attachment import AttachmentTypeEnum
from vectice.api.json.code_version import CodeVersion
from vectice.api.json.dataset_register import DatasetRegisterInput, DatasetRegisterOutput
from vectice.api.json.dataset_representation import DatasetRepresentationOutput
from vectice.api.json.dataset_resource_representation import DatasetResourceRepresentationOutput
from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput
from vectice.api.json.issue import IssueOutput
from vectice.api.json.iteration import (
    IterationContextInput,
    IterationStatus,
    IterationUpdateInput,
    RetrieveIterationOutput,
)
from vectice.api.json.metric import MetricInput
from vectice.api.json.model_representation import ModelRepresentationOutput
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput, ModelVersionUpdateInput
from vectice.api.json.organization_config import OrgConfigOutput
from vectice.api.json.report import ReportOutput
from vectice.api.json.review import ReviewOutput
from vectice.api.project import ProjectApi
from vectice.api.project_template import ProjectTemplateApi
from vectice.api.version import VersionApi
from vectice.api.workspace import WorkspaceApi
from vectice.models.attachment_container import FILE_PATH_DOES_NOT_EXIST_ERROR_MESSAGE
from vectice.models.dataset import Dataset
from vectice.models.iteration import Iteration
from vectice.models.model import Model
from vectice.models.phase import Phase
from vectice.models.representation.attachment import Attachment
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation
from vectice.models.resource.metadata.base import DatasetSourceType, DatasetSourceUsage
from vectice.models.table import Table
from vectice.types.iteration import IterationInput
from vectice.types.phase import PhaseInput
from vectice.types.project import ProjectInput
from vectice.types.version import TVersion
from vectice.utils.api_utils import DEFAULT_NUMBER_OF_ITEMS
from vectice.utils.vectice_ids_regex import WORKSPACE_VID_REG

if TYPE_CHECKING:
    from io import BytesIO, IOBase

    from requests import Response

    from vectice.api.json import ProjectOutput, ProjectTemplateOutput, WorkspaceOutput
    from vectice.api.json.compatibility import CompatibilityOutput
    from vectice.api.json.iteration import (
        IterationOutput,
        IterationStepArtifact,
        IterationStepArtifactInput,
    )
    from vectice.api.json.phase import PhaseOutput
    from vectice.api.json.report import ReportOutput

_logger = logging.getLogger(__name__)


DISABLED_FEATURE_FLAG_MESSAGE = (
    "This '{}' feature is not enabled. Please contact your sales representative for Beta program access."
)


class Client:
    """Low level Vectice API client."""

    _instance: Client | None = None

    def __init__(
        self,
        token: str,
        api_endpoint: str,
    ):
        self.auth = Auth(api_endpoint=api_endpoint, api_token=token)
        transport = RequestsHTTPTransport(url=self.auth.api_base_url + "/graphql", verify=self.auth.verify_certificate)
        logging.getLogger("gql.transport.requests").setLevel("WARNING")
        self._gql_client = GQLClient(transport=transport)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._org_config = self._get_org_config()

    @staticmethod
    def get_instance(
        token: str | None = None,
        api_endpoint: str | None = None,
    ) -> Client:
        if token or api_endpoint:
            Client._instance = None

        if Client._instance is None:
            if token is None or api_endpoint is None:
                raise VecticeException(
                    "You should login before using this method. See https://api-docs.vectice.com/reference/vectice/connection/#vectice.Connection.connect for more info."
                )

            Client._instance = Client(token, api_endpoint)
        return Client._instance

    @property
    def version_api(self) -> str:
        return __version__

    @property
    def version_backend(self) -> str:
        versions = VersionApi(self._gql_client, self.auth).get_public_config().versions
        for version in versions:
            if version.artifact_name == ArtifactName.BACKEND:
                return version.version
        raise ValueError("No version found for backend.")

    def check_compatibility(self) -> CompatibilityOutput:
        return CompatibilityApi(self.auth).check_version()

    @property
    def org_config(self) -> OrgConfigOutput:
        return self._org_config

    def _get_org_config(self) -> OrgConfigOutput:
        return GqlOrganizationApi(self._gql_client, self.auth).get_organization_config()

    def list_projects(
        self,
        workspace: str,
        size: int = DEFAULT_NUMBER_OF_ITEMS,
    ) -> PagedResponse[ProjectOutput]:
        """List the projects in a workspace.

        Parameters:
            workspace: The workspace id.

        Returns:
            The workspace's projects.
        """
        return ProjectApi(self._gql_client, self.auth).list_projects(workspace, size)

    def get_project(self, project: str, workspace: str | None = None) -> ProjectOutput:
        """Get a project.

        Parameters:
            project: The project name or vectice id.
            workspace: The workspace name or id.

        Returns:
            The project JSON structure.
        """
        if workspace is not None and not re.search(WORKSPACE_VID_REG, workspace):
            workspace = WorkspaceApi(self._gql_client, self.auth).get_workspace(workspace).id
        return ProjectApi(self._gql_client, self.auth).get_project(project, workspace)

    def create_project(self, workspace_id: str, project: ProjectInput) -> ProjectOutput:
        return ProjectApi(self._gql_client, self.auth).create_project(workspace_id, project)

    def get_workspace(self, workspace: str) -> WorkspaceOutput:
        """Get a workspace.

        Parameters:
            workspace: The workspace name or id.

        Returns:
            The workspace JSON structure.
        """
        return WorkspaceApi(self._gql_client, self.auth).get_workspace(workspace)

    def list_templates(
        self,
    ) -> list[ProjectTemplateOutput]:
        """List the project templates in a workspace.

        Returns:
            The workspace's project templates.
        """
        return ProjectTemplateApi(self._gql_client, self.auth).list_templates()

    def list_workspaces(self, size: int = DEFAULT_NUMBER_OF_ITEMS) -> PagedResponse[WorkspaceOutput]:
        """List the workspaces.

        Returns:
            The workspaces.
        """
        return WorkspaceApi(self._gql_client, self.auth).list_workspaces(size)

    def upsert_code_attachments(self, files: list[tuple[str, tuple[str, str]]], code_version_id: int):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            code_version_id: The code version id to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files=files, code_version_id=code_version_id)

    def attach_file_to_lineage(self, lineage_id: int, entity_file_id: int):
        """Attach an attachment to an assets lineage.

        Parameters:
            lineage_id: The lineage id to the file to.
            entity_file_id: The file id to attach

        Returns:
            The JSON structure.
        """
        return GqlAttachmentApi(self._gql_client, self.auth).attach_file_to_lineage(
            lineage_id=lineage_id, entity_file_id=entity_file_id
        )

    def upsert_version_attachments(self, files: list[tuple[str, tuple[str, BinaryIO]]], version: TVersion):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            version: The version to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).post_attachment(files, version)

    def upsert_version_tables(self, tables: list[Table], version: TVersion):
        """Upsert a table.

        Parameters:
            tables: The tables to attach.
            version: The version to attach tables to.

        Returns:
            The JSON structure.
        """
        return GqlAttachmentApi(self._gql_client, self.auth).upsert(
            version.id,
            (
                "MODEL_VERSION"
                if isinstance(version, (ModelVersionOutput, ModelVersionRepresentationOutput))
                else "DATASET_VERSION"
            ),
            tables,
        )

    def upsert_iteration_attachments(
        self,
        files: list[tuple[str, tuple[str, BytesIO | IOBase]]],
        iteration_id: str,
        step_name: str | None = None,
    ):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            iteration_id: The iteration id to attach files to.
            step_name (Optional): The step name to add this attachment as an artifact of the step.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files, iteration_id=iteration_id, step_name=step_name)

    def upsert_lineage_attachments(
        self,
        files: list[tuple[str, tuple[str, BytesIO | IOBase]]],
        lineage_id: int,
    ):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            lineage_id: The lineage id to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files, lineage_id=lineage_id)

    def create_model_predictor(self, model_type: str, model_content: BytesIO, model_version: ModelVersionOutput):
        """Create a predictor.

        Parameters:
            model_type: The type of model to attach.
            model_content: The binary content of the model.
            model_version: The model version to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).post_model_predictor(model_type, model_content, model_version)

    def get_version_table(
        self,
        asset: ModelVersionRepresentation | DatasetVersionRepresentation | Iteration,
        table: str,
    ) -> dict[str, Any]:
        """Get a version table.

        Parameters:
            version: The version to get the table from.

        Returns:
            The version table.
        """
        return GqlAttachmentApi(self._gql_client, self.auth).get_version_table(
            asset.id,
            get_asset_type(asset),
            table,
        )

    def list_attachments(
        self, asset: ModelVersionRepresentation | DatasetVersionRepresentation | Iteration
    ) -> list[Attachment]:
        """List the attachments of an asset.

        Parameters:
            asset: The asset to list attachments from.

        Returns:
            The attachments of the asset.
        """
        attachments = GqlAttachmentApi(self._gql_client, self.auth).list_attachments(
            asset.id,
            get_asset_type(asset),
        )

        return list(
            map(
                lambda attachment: Attachment(
                    id=attachment.id,
                    name=attachment.name,
                    type="file" if attachment.type == AttachmentTypeEnum.ENTITY_FILE else "table",
                ),
                attachments,
            )
        )

    def download_attachment(
        self,
        asset: ModelVersionRepresentation | DatasetVersionRepresentation | Iteration,
        attachment: str,
        file_path: str,
    ) -> None:
        """Download the attachment of a version.

        Parameters:
            asset: The asset to download the attachment from.
            attachment: The attachment to download.
            file_path: The path to download the attachment to.

        Returns:
            None.
        """
        return AttachmentApi(self.auth).download_attachment(asset, attachment, file_path)

    def get_code_version_attachment(self, code_version_id: int, file_id: int) -> Response:
        """Get the attachment of a code version.

        Parameters:
            code_version_id: The code version id to list attachments from.
            file_id: The file id attached to the code version.

        Returns:
            The file attached to the code version.
        """
        return AttachmentApi(self.auth).get_code_version_attachment(code_version_id=code_version_id, file_id=file_id)

    def list_phases(self, project: str, size: int = DEFAULT_NUMBER_OF_ITEMS) -> PagedResponse[PhaseOutput]:
        return GqlPhaseApi(self._gql_client, self.auth).list_phases(project, size)

    def get_phase(self, phase: str, project_id: str | None = None) -> PhaseOutput:
        if project_id is None:
            raise MissingReferenceError("project")
        return GqlPhaseApi(self._gql_client, self.auth).get_phase(phase, project_id)

    def create_phase(self, project_id: str, phase: PhaseInput) -> PhaseOutput:
        return GqlPhaseApi(self._gql_client, self.auth).create_phase(project_id, phase)

    def get_full_phase(self, phase: str) -> PhaseOutput:
        return GqlPhaseApi(self._gql_client, self.auth).get_phase(phase=phase, full=True)

    def list_sections(self, iteration_id: str) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).list_sections(iteration_id)

    def list_reviews(self, phase: str, size: int = DEFAULT_NUMBER_OF_ITEMS) -> PagedResponse[ReviewOutput]:
        return GqlReviewApi(self._gql_client, self.auth).list_reviews(phase, size)

    def list_resources_items(
        self, id: str, type: DatasetSourceUsage | None
    ) -> list[DatasetResourceRepresentationOutput]:
        return GqlDatasetApi(self._gql_client, self.auth).get_resources_items(id, type)

    def list_columns(self, id: int, type: DatasetSourceType) -> list[Dict[str, Any]]:
        return GqlDatasetApi(self._gql_client, self.auth).get_columns(id, type)

    def list_iterations(
        self,
        phase: str,
        only_mine: bool = False,
        statuses: list[IterationStatus] | None = None,
        size: int = DEFAULT_NUMBER_OF_ITEMS,
    ) -> PagedResponse[IterationOutput]:
        return IterationApi(self._gql_client, self.auth).list_iterations(phase, only_mine, statuses, size)

    def list_iteration_assets(self, iteration: str) -> PagedResponse[IterationStepArtifact]:
        return IterationApi(self._gql_client, self.auth).list_iteration_assets(iteration)

    def list_step_definitions(self, phase: str) -> PagedResponse[RequirementOutput]:
        return GqlPhaseApi(self._gql_client, self.auth).list_step_definitions(phase)

    def get_last_iteration(self, phase_id: str, iteration: IterationInput | None = None) -> RetrieveIterationOutput:
        return IterationApi(self._gql_client, self.auth).get_last_iteration(phase_id, iteration)

    def get_active_iteration_or_create(self, phase_id: str) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_active_iteration_or_create(phase_id)

    def get_iteration_by_id(self, iteration_id: str, full: bool = False) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_iteration_by_id(iteration_id, full)

    def get_iteration_by_index(self, phase_id: str, index: int) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_iteration_by_index(phase_id, index)

    def create_iteration(self, phase_id: str, iteration: IterationInput | None = None) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).create_iteration(phase_id, iteration)

    def update_iteration(self, iteration_id: str, iteration: IterationUpdateInput) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).update_iteration(iteration, iteration_id)

    def delete_iteration(self, iteration_id: str) -> None:
        IterationApi(self._gql_client, self.auth).delete_iteration(iteration_id)

    def add_iteration_artifacts(
        self, iteration_id: str, artifacts: list[IterationStepArtifactInput], section: str | None = None
    ) -> list[IterationStepArtifact]:
        iteration_context = IterationContextInput(iterationId=iteration_id, stepName=section)
        return IterationApi(self._gql_client, self.auth).add_iteration_artifacts(
            artifacts=artifacts, iteration_context=iteration_context
        )

    def remove_iteration_assets(self, iteration_id: str, section: str | None = None, all: bool = False):
        iteration_context = IterationContextInput(iterationId=iteration_id, stepName=section)
        return IterationApi(self._gql_client, self.auth).remove_iteration_assets(
            iteration_context=iteration_context, all=all
        )

    def remove_assets_from_iteration(self, iteration_id: str, artifacts_id_list: list[int]):
        return IterationApi(self._gql_client, self.auth).remove_assets_from_iteration(iteration_id, artifacts_id_list)

    def save_autolog_cells(
        self, iteration_id: str, cells: list[dict[str, str]], prefix: str | None = None, is_trace: bool = False
    ):
        return IterationApi(self._gql_client, self.auth).save_autolog_cells(iteration_id, cells, prefix, is_trace)

    def organize_with_ai(self, iteration_id: str):
        return IterationApi(self._gql_client, self.auth).organize_with_ai(iteration_id)

    def organize_with_ai_and_report(self, iteration_id: str) -> ReportOutput:
        return IterationApi(self._gql_client, self.auth).organize_with_ai_and_report(iteration_id)

    def register_dataset_from_source(
        self,
        dataset: Dataset,
        iteration_id: str,
        code_version: CodeVersion | None = None,
        section: str | None = None,
    ) -> DatasetRegisterOutput:
        name = self.get_dataset_name(dataset)
        derived_from = self.get_derived_from(dataset)
        metadata_asdict, is_updating_columns = dataset._metadata_as_dict, dataset._is_updating_columns  # type: ignore
        properties, _ = self.get_properties_and_metrics(dataset)

        dataset_register_input = DatasetRegisterInput(
            name=name,
            type=dataset.type.value,
            datasetSources=metadata_asdict,
            datasetInputs=[s for s in derived_from if s.startswith("DTV-")] if derived_from else None,
            modelInputs=[s for s in derived_from if s.startswith("MDV-")] if derived_from else None,
            codeVersion=code_version.asdict() if code_version is not None else None,
            properties=properties,
        )

        iteration_context = IterationContextInput(iterationId=iteration_id, stepName=section)
        dataset_register_output = self.register_dataset(dataset_register_input, iteration_context=iteration_context)
        dtv_id = dataset_register_output["datasetVersion"]["vecticeId"]

        if is_updating_columns:
            self.warn_if_dataset_version_columns_are_missing_description(dtv_id)

        dataset.latest_version_id = dtv_id
        return dataset_register_output

    def warn_if_dataset_version_columns_are_missing_description(self, dataset_version_id: str):
        has_columns_without_description = GqlDatasetApi(
            self._gql_client, self.auth
        ).get_dataset_version_has_columns_without_description(dataset_version_id)
        if has_columns_without_description is True:
            _logger.warning(
                "Some column descriptions are missing. Consider adding descriptions for improved data clarity."
            )

    def get_dataset(self, id: str) -> DatasetRepresentationOutput:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset(id)

    def get_dataset_version(self, id: str) -> DatasetVersionRepresentationOutput:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset_version(id)

    def update_dataset_version(self, id: str, columns_description: list[dict[str, str]]):
        return GqlDatasetApi(self._gql_client, self.auth).update_dataset_version(id, columns_description)

    def update_columns_description_via_csv(self, id: str, file_path: str):
        if not os.path.exists(file_path):
            raise ValueError(FILE_PATH_DOES_NOT_EXIST_ERROR_MESSAGE % file_path)
        curr_file = ("file", (file_path, open(file_path, "rb")))
        ColumnsDescriptionApi(self.auth).post_columns_description(id, curr_file)
        curr_file[1][1].close()

    def get_model(self, id: str) -> ModelRepresentationOutput:
        return GqlModelApi(self._gql_client, self.auth).get_model(id)

    def get_dataset_list(
        self, project_id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[DatasetRepresentationOutput]:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset_list(project_id, size)

    def get_dataset_version_list(
        self, id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[DatasetVersionRepresentationOutput]:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset_version_list(id, size)

    def get_model_list(
        self, project_id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[ModelRepresentationOutput]:
        return GqlModelApi(self._gql_client, self.auth).get_model_list(project_id, size)

    def get_model_version_list(
        self, id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[ModelVersionRepresentationOutput]:
        return GqlModelApi(self._gql_client, self.auth).get_model_version_list(id, size)

    def get_model_version(self, id: str) -> ModelVersionRepresentationOutput:
        return GqlModelApi(self._gql_client, self.auth).get_model_version(id)

    def get_model_version_approval_history(self, id: str) -> dict[str, datetime | None]:
        response = GqlModelApi(self._gql_client, self.auth).get_model_version_approval_history(id)

        def _transform_dict(acc: dict[str, datetime | None], item: tuple[str, str | None]):
            key, value = item
            camel_case_key = re.sub(r"(?<!^)(?=[A-Z])", "_", key).lower()
            return {**acc, camel_case_key: read_nodejs_date(value) if value else None}

        return functools.reduce(_transform_dict, list(response.items()), {})

    def get_lineage_inputs(
        self, id: str, type: Literal["mdv"] | Literal["dtv"]
    ) -> list[DatasetVersionRepresentationOutput]:
        return GqlLineageApi(self._gql_client, self.auth).get_lineage_inputs(id, type)

    def get_lineage_children(
        self, id: str, type: Literal["mdv"] | Literal["dtv"]
    ) -> tuple[list[ModelVersionRepresentationOutput], list[DatasetVersionRepresentationOutput]]:
        return GqlLineageApi(self._gql_client, self.auth).get_lineage_children(id, type)

    def get_reports(
        self, id: str, type: Literal["mdv"] | Literal["prj"], size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[ReportOutput]:
        return GqlReportApi(self._gql_client, self.auth).get_reports(id, type, size)

    def get_issues(
        self, id: str, type: Literal["mdv"] | Literal["prj"], size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[IssueOutput]:
        return GqlIssueApi(self._gql_client, self.auth).get_issues(id, type, size)

    def update_model(self, model_version_id: str, model: ModelVersionUpdateInput):
        GqlModelApi(self._gql_client, self.auth).update_model(model_version_id, model)

    def has_column(self, id: str, column: str) -> bool:
        return GqlDatasetApi(self._gql_client, self.auth).has_column(id, column)

    @staticmethod
    def get_dataset_name(dataset: Dataset) -> str:
        return (
            f"dataset {datetime.time}"
            if dataset.name is None  # pyright: ignore[reportUnnecessaryComparison]
            else dataset.name
        )

    @staticmethod
    def get_derived_from(obj: Dataset | Model) -> list[str] | None:
        return None if obj.derived_from is None else obj.derived_from  # pyright: ignore[reportUnnecessaryComparison]

    @staticmethod
    def get_properties_and_metrics(
        obj: Dataset | Model,
    ) -> tuple[list[dict[str, Any]] | None, list[dict[str, Any]] | None]:
        properties = [vars(PropertyInput(prop.key, prop.value)) for prop in obj.properties] if obj.properties else None
        metrics = (
            [vars(MetricInput(met.key, met.value)) for met in obj.metrics]
            if (isinstance(obj, Model) and obj.metrics)
            else None
        )
        return properties, metrics

    def register_dataset(
        self,
        dataset_register_input: DatasetRegisterInput,
        iteration_context: IterationContextInput,
    ) -> DatasetRegisterOutput:
        data: DatasetRegisterOutput = GqlDatasetApi(self._gql_client, self.auth).register_dataset(
            dataset_register_input, iteration_context
        )
        _logger.debug(
            f"Successfully registered Dataset("
            f"name='{dataset_register_input.name}', "
            f"id={data['datasetVersion']['vecticeId']}, "
            f"version='{data['datasetVersion']['name']}', "
            f"type={dataset_register_input.type})."
        )
        return data

    def register_model(
        self,
        model: Model,
        iteration: Iteration,
        phase: Phase,
        code_version: CodeVersion | None = None,
        section: str | None = None,
    ) -> ModelRegisterOutput:
        """Register a model.

        Parameters:
            model: The model to register
            iteration: The iteration
            phase: The phase
            code_version_id: The code version ID
            section: The section

        Returns:
            The registered model.
        """
        properties, metrics = self.get_properties_and_metrics(model)
        derived_from = self.get_derived_from(model)
        model.name = (model.name or f"{phase.name} {iteration.index} model")[:60]
        model_register_input = ModelRegisterInput(
            name=model.name,
            modelType="OTHER",
            status=ModelVersionStatus.EXPERIMENTATION.value,
            datasetInputs=[s for s in derived_from if s.startswith("DTV-")] if derived_from else None,
            modelInputs=[s for s in derived_from if s.startswith("MDV-")] if derived_from else None,
            metrics=metrics,
            properties=properties,
            algorithmName=model.technique,
            framework=model.library,
            codeVersion=code_version.asdict() if code_version is not None else None,
            context=model.additional_info.asdict() if model.additional_info is not None else None,
        )
        iteration_context = IterationContextInput(iterationId=iteration.id, stepName=section)

        model_output = GqlModelApi(self._gql_client, self.auth).register_model(
            model_register_input, iteration_context=iteration_context
        )
        model.latest_version_id = model_output["modelVersion"]["vecticeId"]
        return model_output

    def upsert_properties(self, type: str, id: str, properties: list[dict[str, str | int]]):
        return GqlPropertyApi(self._gql_client, self.auth).upsert(type, id, properties)

    def upsert_metrics(self, type: str, id: str, metrics: list[dict[str, str | float | int]]):
        return GqlMetricApi(self._gql_client, self.auth).upsert(type, id, metrics)

    def get_user_and_default_workspace(self):
        return UserAndDefaultWorkspaceApi(self._gql_client, self.auth).get_user_and_default_workspace()

    def assert_feature_flag_or_raise(self, code: str) -> None:
        enabled = self.is_feature_flag_enabled(code)
        if enabled is False:
            raise VecticeException(DISABLED_FEATURE_FLAG_MESSAGE.format(code))

    def is_feature_flag_enabled(self, code: str) -> bool:
        enabled = GqlFeatureFlagApi(self._gql_client, self.auth).is_feature_flag_enabled(code)
        if enabled is False:
            _logger.info(DISABLED_FEATURE_FLAG_MESSAGE.format(code))
        return enabled

    def is_feature_flag_enabled_without_message(self, code: str) -> bool:
        return GqlFeatureFlagApi(self._gql_client, self.auth).is_feature_flag_enabled(code)



from enum import Enum


class ModelLibrary(Enum):
    """Enumeration that defines what the model library."""

    SKLEARN = "SKLEARN"
    SKLEARN_PIPELINE = "SKLEARN_PIPELINE"
    SKLEARN_SEARCH = "SKLEARN_SEARCH"
    LIGHTGBM = "LIGHTGBM"
    CATBOOST = "CATBOOST"
    KERAS = "KERAS"
    STATSMODEL = "STATSMODEL"
    PYTORCH = "PYTORCH"
    PYSPARKML = "PYSPARKML"
    H2O = "H2O"
    NONE = "NONE"



from __future__ import annotations

from importlib.util import find_spec
from typing import Any

from modeva import DataSet, FactSheet  # pyright: ignore[reportMissingImports]
from modeva.models.wrappers.api import (  # pyright: ignore[reportMissingImports]
    modeva_arbitrary_classifier,
    modeva_arbitrary_regressor,
    modeva_sklearn_classifier,
    modeva_sklearn_regressor,
)
from modeva.utils.results import ValidationResult  # type: ignore[reportMissingImports]
from pandas import DataFrame

SKLEARN_MODEL_WRAPPERS = [modeva_sklearn_classifier, modeva_sklearn_regressor]
ARBITARY_MODEL_WRAPPERS = [modeva_arbitrary_classifier, modeva_arbitrary_regressor]

is_selenium = find_spec("selenium") is not None
if is_selenium:
    from functools import wraps

    from selenium.webdriver.chrome.options import Options  # type: ignore[reportMissingImports]

    original_add_argument = Options.add_argument

    def add_argument_with_extras(func):  # type: ignore[reportMissingParameterType]
        @wraps(func)
        def wrapper(self, argument):  # type: ignore[reportMissingParameterType]
            # call the original method
            result = func(self, argument)
            # add extra options
            func(self, "--disable-dev-shm-usage")
            func(self, "--disable-gpu")
            func(self, "--headless")
            func(self, "--no-sandbox")
            return result

        return wrapper

    Options.add_argument = add_argument_with_extras(Options.add_argument)


def modeva_wrapper(
    dataset_train: DataFrame,
    dataset_test: DataFrame,
    main_model: Any | None = None,
    comparison_models: list | None = None,
    dataset_out: DataFrame | None = None,
    target: str | None = None,
) -> FactSheet:
    """Wraps the input datasets and models into a Vectice fact sheet.
       You can pass either a main_model or comparison_models.

    Args:
        dataset_train (DataFrame): The training dataset used for model evaluation.
        dataset_test (DataFrame): The test dataset used for model evaluation.
        main_model (Any): The main model to be evaluated.
        comparison_models (list): A list of models to be evaluated.
        dataset_out (DataFrame, optional): An optional out of time dataset. Defaults to None.
        target (str, optional): The target column of the dataset_train provided. Defaults to the last column.

    Returns:
        FactSheet: A fact sheet object that contains details about the datasets and models.
    """
    fact_sheet = VecticeFactSheet(dataset_train, dataset_test, main_model, comparison_models, dataset_out, target)
    return fact_sheet


class VecticeFactSheet(FactSheet):  # type: ignore[reportUntypedBaseClass]
    def __init__(
        self,
        dataset_train: DataFrame,
        dataset_test: DataFrame,
        main_model: Any | None = None,
        comparison_models: list | None = None,
        dataset_out: DataFrame | None = None,
        target: str | None = None,
    ):
        self._dataset = self._dataset_setup(dataset_train, dataset_test, dataset_out, target)
        if not comparison_models:
            self._main_model = self._get_sklearn_model(main_model) or self._get_arbitrary_model(main_model)
        else:
            self._main_model = None
        self._comparison_models = self._models_setup(comparison_models) if comparison_models else None
        self._validation_test_results: list[ValidationResult] | list = []

        super().__init__(self._dataset, self._main_model, self._comparison_models)

    def run_robustness_test_suite(self):
        from tqdm.notebook import tqdm  # pyright: ignore[reportMissingImports]

        # robustness_tests = [self.explain_shap, self.diagnose_accuracy_table, self.diagnose_robustness, self.diagnose_reliability, self.diagnose_resilience]
        robustness_tests = [self.explain_shap, self.diagnose_accuracy_table, self.diagnose_resilience]

        with tqdm(robustness_tests) as pbar:
            for test in robustness_tests:
                pbar.set_description(f"Test Suite {test.__name__} running")
                result = test()
                pbar.update(1)
                self._validation_test_results.append(result)

    def _get_arbitrary_model(self, model: Any):
        # try arb
        for wrapper in ARBITARY_MODEL_WRAPPERS:
            algorithmn_name = str(model.__class__).split(".")[-1]
            algorithmn_name = algorithmn_name.replace("'>", "")
            try:
                # regressor / params are for regression models
                wrapped = wrapper(name=algorithmn_name, predict_function=model.predict, fit_function=model.fit)
                self._dataset.set_task_type("Regression")
                return wrapped
            except Exception:
                pass
            try:
                # classifier / params are for classification models
                wrapped = wrapper(
                    name=algorithmn_name,
                    predict_function=model.predict,
                    predict_proba_function=model.predict_proba,
                    fit_function=model.fit,
                )
                self._dataset.set_task_type("Classification")
                return wrapped
            except Exception:
                pass

    def _get_sklearn_model(self, model: Any):
        from sklearn.base import is_classifier, is_regressor

        for wrapper in SKLEARN_MODEL_WRAPPERS:
            algorithmn_name = str(model.__class__).split(".")[-1]
            algorithmn_name = algorithmn_name.replace("'>", "")
            try:
                wrapped = wrapper(name=algorithmn_name, estimator=model)
                if is_regressor(model):
                    self._dataset.set_task_type("Regression")
                if is_classifier(model):
                    self._dataset.set_task_type("Classification")
                return wrapped
            except Exception:
                pass

    def _models_setup(self, models: list) -> list:
        # converts estimators to modeva wrapped models
        wrapped_models = []
        for model in models:
            # modeva wrappers seem to generalize well for the following
            # sklearn, xgboost, lightgbm and catboost
            wrapped_model = self._get_sklearn_model(model) or self._get_arbitrary_model(model)
            if wrapped_model:
                wrapped_models.append(wrapped_model)
        return wrapped_models

    def _dataset_setup(
        self,
        dataset_train: DataFrame,
        dataset_test: DataFrame,
        dataset_out: DataFrame | None = None,
        target: str | None = None,
    ) -> DataSet:
        # TODO series data / numpy arrays
        # takes train & test and creates a modeva DataSet
        dataset = DataSet()
        dataset.load_dataframe_train_test(train=dataset_train, test=dataset_test)
        if dataset_out is not None:
            dataset.set_raw_extra_data(name="oot", data=dataset_out)
        # TODO add a target parameter
        # assume initially
        if target is None:
            target = dataset_train.columns[-1]
        dataset.set_target(target)
        return dataset



from __future__ import annotations

import logging
import os
from io import BytesIO
from typing import TYPE_CHECKING, Any, Dict, cast

import pandas as pd
import requests
from matplotlib import pyplot as plt
from PIL import Image

from vectice.models import AdditionalInfo, Framework
from vectice.models.dataset import TDerivedFrom
from vectice.models.model_exp_tracker import ModelExpTracker
from vectice.utils.common_utils import check_image_path

_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import wandb
    from wandb.apis.public.runs import Run


class WandbModel(ModelExpTracker):
    def __init__(
        self,
        run_id: str,
        client: wandb.Api | None = None,
        url: str | None = None,
        derived_from: list[TDerivedFrom] | None = None,
        entity: str | None = None,
        project: str | None = None,
    ):
        """Args:
        run_id: Either 'entity/project/run_id' or just the run hash.
        client: Optional wandb.Api client. If not provided, a new one is created.
        url: Optional public or internal URL to the run.
        derived_from: Optional lineage.
        entity: Required if using short run_id.
        project: Required if using short run_id.
        """
        if client is None:
            client = wandb.Api()

        if run_id.count("/") == 2:
            run_path = run_id
        elif entity and project:
            run_path = f"{entity}/{project}/{run_id}"
        else:
            raise ValueError(
                "If 'run_id' is not in 'entity/project/run_id' format, you must also provide 'entity' and 'project'."
            )

        super().__init__(run_id=run_path, client=client, derived_from=derived_from)
        self.url = url
        self._run: Run = self.client.run(run_path)

    def get_metrics(self) -> Dict[str, Any]:
        excluded_keys = {"_step", "_runtime", "_timestamp", "_wandb"}
        metrics = {
            str(key): value
            for key, value in self._run.summary._json_dict.items()  # pyright: ignore[reportPrivateUsage]
            if isinstance(value, (int, float)) and key not in excluded_keys
        }
        return metrics

    def get_properties(self) -> Dict[str, Any]:
        props = dict(self._run.config)

        props.update(
            {
                "run_id": self._run.id,
                "state": self._run.state,
                "created_at": str(self._run.created_at),
                "project": self._run.project,
                "entity": self._run.entity,
                "tags": ", ".join(self._run.tags) if self._run.tags else None,
            }
        )

        if hasattr(self._run, "commit"):
            props["git_commit"] = self._run.commit
        if hasattr(self._run, "repo") and self._run.repo:
            props["git_repo_url"] = self._run.repo.get("repo_url")
            props["git_branch"] = self._run.repo.get("branch")
            props["git_tag"] = self._run.repo.get("tag")

        if self._run.sweep:
            props["sweep_id"] = self._run.sweep.id
            props["sweep_name"] = self._run.sweep.name
            props["sweep_config"] = str(self._run.sweep.config)

        return props

    def get_attachments(self) -> list[str] | None:
        try:

            output_dir = "wandb_attachments"
            os.makedirs(output_dir, exist_ok=True)
            image_paths = []

            # 1. Download image and media files
            for file in self._run.files():
                if file.name.lower().endswith((".png", ".jpg", ".jpeg", ".webp", ".svg")):
                    downloaded = file.download(root=output_dir, replace=True)
                    if check_image_path(downloaded.name):
                        image_paths.append(downloaded.name)

            # 2. Download media images (e.g., wandb.Image(...))
            for file in self._run.files():
                if file.name.startswith("media/images/"):
                    try:
                        response = requests.get(file.url, stream=True)
                        response.raise_for_status()
                        image = Image.open(BytesIO(response.content)).convert("RGB")
                        base_name = os.path.splitext(os.path.basename(file.name))[0]
                        save_path = os.path.join(output_dir, f"{base_name}.png")
                        image.save(save_path, format="PNG")
                        image_paths.append(save_path)
                    except Exception:
                        continue  # skip broken media

            # 3. Regenerate history plots per metric
            try:
                history_df = cast(pd.DataFrame, self._run.history(samples=1000, pandas=True))
                excluded_keys = {"_step", "_runtime", "_timestamp", "_wandb", "step"}

                metric_columns = [
                    col
                    for col in history_df.columns
                    if col not in excluded_keys and history_df[col].dtype.kind in {"i", "f"}
                ]

                for metric in metric_columns:
                    series = history_df[metric].dropna()
                    if series.empty:
                        continue

                    fig, ax = plt.subplots()
                    ax.plot(series.index, series.values)  # pyright: ignore
                    ax.set_title(metric)
                    ax.set_xlabel("Index (step/epoch)")
                    ax.set_ylabel("Value")
                    safe_name = metric.replace("/", "_")
                    chart_path = os.path.join(output_dir, f"{safe_name}.png")
                    fig.savefig(chart_path)
                    plt.close(fig)
                    image_paths.append(chart_path)

            except Exception as e:
                _logger.warning(f"Could not regenerate charts from history: {e}")

            return image_paths

        except Exception as e:
            _logger.warning(f"Could not extract W&B images: {e}")
            return None

    def get_name(self) -> str | None:
        return self._run.name

    def get_additional_info(self) -> AdditionalInfo:
        return AdditionalInfo(url=self.url, run=self.run, framework=Framework.WANDB)



from __future__ import annotations

from typing import TypedDict, TypeVar

from vectice.autolog.asset_services.service_types.vectice_types import VecticeType
from vectice.models.dataset import Dataset
from vectice.models.model import Model
from vectice.models.representation.dataset_representation import DatasetRepresentation
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_representation import ModelRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation
from vectice.models.table import Table
from vectice.models.validation import ValidationModel

VecticeObjectTypes = TypeVar(
    "VecticeObjectTypes",
    Dataset,
    Model,
    Table,
    ValidationModel,
    DatasetVersionRepresentation,
    ModelVersionRepresentation,
    ModelRepresentation,
    DatasetRepresentation,
)
TVecticeObjects = TypedDict(
    "TVecticeObjects", {"variable": str, "vectice_object": VecticeObjectTypes, "asset_type": VecticeType}
)
VecticeObjectClasses = (
    Dataset,
    Model,
    Table,
    ValidationModel,
    DatasetVersionRepresentation,
    ModelVersionRepresentation,
    ModelRepresentation,
    DatasetRepresentation,
)


class AutologVecticeAssetService:
    def __init__(self, key: str, asset: VecticeObjectTypes):
        self._asset = asset
        self._key = key

    def get_asset(self):
        return {"variable": self._key, "vectice_object": self._asset, "asset_type": VecticeType.VECTICE_OBJECT}



from __future__ import annotations

import glob
import hashlib
import logging
import os
from datetime import datetime, timezone

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata.base import DatasetSourceOrigin
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.files_metadata import File, FilesMetadata

_logger = logging.getLogger(__name__)

TDataFrameType = DataFrameType


class FileResource(Resource):
    """Wrap columnar data and its metadata in a local file.

    This resource wraps data that you have stored in a local file with optional metadata and versioning.
    You pass it as an argument of your Vectice Dataset wrapper before logging it to an iteration.

    ```python
    from vectice import FileResource

    resource = FileResource(paths="my/file/path")
    ```

    NOTE: **Vectice does not store your actual dataset. Instead, it stores your dataset's metadata, which is information about your dataset.
    These details are captured by resources tailored to the specific environment in use, such as: local (FileResource), Bigquery, S3, GCS...**
    """

    _origin = DatasetSourceOrigin.LOCAL.value

    def __init__(
        self,
        paths: str | list[str],
        dataframes: TDataFrameType | list[TDataFrameType] | None = None,
        capture_schema_only: bool = False,
        columns_description: dict[str, str] | str | None = None,
    ):
        """Initialize a file resource.

        Parameters:
            paths: The paths of the files to wrap.
            dataframes (Optional): The dataframes allowing vectice to optionally compute more metadata about this resource such as columns stats. (Support Pandas, Spark, H2O)
            capture_schema_only (Optional): A boolean parameter indicating whether to capture only the schema or both the schema and column statistics of the dataframes.
            columns_description (Optional): A dictionary or path to a csv file to map the column's name to a specific description. Dictionary should follow the format { "column_name": "Description", ... }

        Examples:
            The following example shows how to wrap a CSV file
            called `iris.csv` in the current directory:

            ```python
            from vectice import FileResource
            iris_trainset = FileResource(paths="iris.csv")
            ```
        """
        super().__init__(
            paths=paths,
            dataframes=dataframes,
            capture_schema_only=capture_schema_only,
            columns_description=columns_description,
        )
        paths_log = ", ".join(self._paths) if len(self._paths) > 1 else self._paths[0]
        _logger.info(f"File: {paths_log} wrapped successfully.")

    def _fetch_data(self) -> dict[str, bytes]:
        datas = {}
        metadata: FilesMetadata = self.metadata  # type:ignore[assignment]
        for file in metadata.files:
            with open(file.name, "rb") as opened_file:
                datas[file.name] = opened_file.read()
        return datas

    def _build_metadata(self) -> FilesMetadata:
        if self._paths is None:  # pyright: ignore[reportUnnecessaryComparison]
            raise ValueError("Paths can not be None.")
        size = None
        files = []
        df_index = 0
        for path in self._paths:
            new_files, total_size, new_df_index = self._file_visitor_list_builder(index=df_index, path=path)
            df_index += new_df_index
            if size is None:
                size = 0
            size += total_size
            files.extend(new_files)

        metadata = FilesMetadata(size=size, origin=self._origin, files=files)
        return metadata

    def _file_visitor_list_builder(self, index: int, path: str) -> tuple[list[File], int, int]:
        files: list[File] = []
        total_size = 0
        df_index = 0
        if not os.path.exists(path):
            raise FileNotFoundError
        if os.path.isfile(path):
            dataframe = (
                self._dataframes[index] if self._dataframes is not None and len(self._dataframes) > index else None
            )
            file: File = self._build_file_from_path(path=path, dataframe=dataframe)
            total_size += file.size or 0
            files.append(file)
            df_index += 1
        else:
            dir_paths = glob.glob(f"{path}/**", recursive=True)
            for entry_path in filter(lambda x: os.path.isfile(x), sorted(dir_paths, key=str.lower)):
                new_index = index + df_index
                dataframe = (
                    self._dataframes[new_index]
                    if self._dataframes is not None and len(self._dataframes) > new_index
                    else None
                )
                glob_file: File = self._build_file_from_path(path=entry_path, dataframe=dataframe)
                total_size += glob_file.size or 0
                files.append(glob_file)
                df_index += 1
        return files, int(total_size), df_index

    def _build_file_from_path(self, path: str, dataframe: TDataFrameType | None = None) -> File:
        entry_stats: os.stat_result = os.stat(path)
        name = path.split("\\")[-1]
        return File(
            name=name,
            size=entry_stats.st_size,
            fingerprint=self._compute_digest_for_path(path),
            updated_date=self._convert_date_to_iso(entry_stats.st_mtime),
            uri=f"file://{os.path.abspath(path)}",
            dataframe=dataframe,
            display_name=name.rpartition("/")[-1],
            capture_schema_only=self.capture_schema_only,
        )

    @staticmethod
    def _compute_digest_for_path(path: str) -> str:
        sha = hashlib.sha256()
        bytes_array = bytearray(128 * 1024)
        mv = memoryview(bytes_array)
        with open(path, "rb", buffering=0) as file:
            n = file.readinto(mv)
            while n:
                sha.update(mv[:n])
                n = file.readinto(mv)
        return sha.hexdigest()

    @staticmethod
    def _convert_date_to_iso(timestamp: float) -> str:
        return datetime.fromtimestamp(timestamp, timezone.utc).isoformat()



from __future__ import annotations

from typing import Any


class CompatibilityOutput:
    def __init__(self, message: str, status: str, *args: Any, **kwargs: Any):
        self._message = message
        self._status = status

    @property
    def message(self):
        return self._message

    @property
    def status(self):
        return self._status



from typing import Union

from pandas.core.frame import DataFrame as PandasDF

from vectice.models.resource.metadata.df_wrapper_resource import DataFrameWrapper

try:
    from pyspark.sql.connect.dataframe import DataFrame as SparkConnectDF
    from pyspark.sql.dataframe import DataFrame as SparkDF

    try:
        from pyspark.pandas.frame import DataFrame as PysparkPandasDF

        DataFramePandasType = Union[PysparkPandasDF, PandasDF]  # type: ignore
    except Exception:
        DataFramePandasType = PandasDF  # type: ignore

    DataFrameTypeWithoutWrapper = Union[DataFramePandasType, SparkDF, SparkConnectDF]  # type: ignore
    DataFrameType = Union[DataFrameTypeWithoutWrapper, DataFrameWrapper]  # type: ignore

except ImportError:
    DataFramePandasType = PandasDF  # type: ignore
    DataFrameType = Union[DataFramePandasType, DataFrameWrapper]  # type: ignore


try:
    from h2o import H2OFrame  # type: ignore[reportMissingImports]

    DataFrameType = Union[DataFrameType, H2OFrame]  # type: ignore[reportMissingImports]
except ImportError:
    pass  # type: ignore

MIN_ROWS_CAPTURE_STATS = 100
ROWS_SAMPLE_CAPTURE_STATS = 400
MAX_COLUMNS_CAPTURE_STATS = 400



from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:

    from catboost.core import CatBoost
    from keras import Model as KerasModel  # type: ignore[reportMissingImports]  # type: ignore[reportMissingImports]
    from keras.models import Model as KerasModel  # type: ignore[reportMissingImports]
    from lightgbm.basic import Booster
    from sklearn.base import BaseEstimator
    from sklearn.pipeline import Pipeline
    from torch.nn import Module as TorchModel

    ModelTypes = TypeVar("ModelTypes", BaseEstimator, Booster, CatBoost, KerasModel, TorchModel, Pipeline)



from __future__ import annotations

import logging
import logging.config
import sys
from collections import defaultdict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vectice.api.json.iteration import IterationStatus
    from vectice.api.json.phase import PhaseStatus


CONNECTION_LOGGING = """Welcome, {user}. You`re now successfully connected to Vectice.

To connect to a phase, copy your phase ID from your Vectice app and use connect.phase("PHA-XXXX")
To create your first iteration do iteration = connect.phase("PHA-XXXX").create_or_get_current_iteration()

If you are using a notebook you can call the help by using a Vectice returned object with the builtin notebook "?":
>> connection?

If you are using an IDE you can call the help() method on any object returned by Vectice:
>> help(connection)

{logging_output}"""

CONNECTION_WORKSPACE_LOGGING = """Welcome, {user!r}. You`re now successfully connected to the workspace {workspace_name!r} in Vectice.

To access a specific project, use \033[1mworkspace\033[0m.project(Project ID)
To get a list of projects you can access and their IDs, use \033[1mworkspace\033[0m.list_projects()

For quick access to the list of projects in the Vectice web app, visit:
{url}/browse/workspace/{workspace_id}"""

CONNECTION_PROJECT_LOGGING = """Welcome, {user!r}. You`re now successfully connected to the project {project_name!r} in Vectice.

To access a specific phase, use \033[1mproject\033[0m.phase(Phase ID)
To get a list of phases you can access and their IDs, use \033[1mproject\033[0m.list_phases()

For quick access to the list of phases in the Vectice web app, visit:
{url}/browse/project/{project_id}"""


class VecticeLoggingStream:
    """A Python stream for use with event logging APIs throughout vectice (`logger.info()`, etc.).

    This stream wraps `sys.stderr`, forwarding `write()`
    and `flush()` calls to the stream referred to by `sys.stderr` at the time of the call.
    It also provides capabilities for disabling the stream to silence event logs and
    enable propagation for pytest.
    """

    def __init__(self):
        self._enabled: bool = True

    def write(self, text: str) -> None:
        if self._enabled:
            sys.stderr.write(text)

    def flush(self):
        if self._enabled:
            sys.stderr.flush()

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._enabled = value


VECTICE_LOGGING_STREAM = VecticeLoggingStream()


def disable_logging():
    """Disable logging.

    Disable the `VecticeLoggingStream` used by event logging APIs throughout Vectice
    `logger.info()` silencing all subsequent event logs.
    """
    VECTICE_LOGGING_STREAM.enabled = False


def enable_logging():
    """Enable logging.

    Enable the `VecticeLoggingStream` used by event logging APIs throughout Vectice.
    This reverses the effects of `disable_logging()`.
    """
    VECTICE_LOGGING_STREAM.enabled = True


def enable_propagation():
    """Enable propagation.

    Enable the `VecticeLoggingStream` propagation used by event logging APIs throughout Vectice.
    The testing suite can use caplog to test the logging stdout.
    """
    configure_vectice_loggers("vectice", propagate=True)


def configure_vectice_loggers(root_module_name: str, propagate: bool = False):
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "vectice_formatter": {
                    "format": None,
                    "datefmt": None,
                },
            },
            "handlers": {
                "vectice_handler": {
                    "formatter": "vectice_formatter",
                    "class": "logging.StreamHandler",
                    "stream": VECTICE_LOGGING_STREAM,
                },
            },
            "loggers": {
                root_module_name: {
                    "handlers": ["vectice_handler"],
                    "level": "INFO",
                    "propagate": propagate,
                },
            },
        }
    )


def format_description(description: str | None) -> str | None:
    if not description:
        return None
    return description


def get_iteration_status(status: IterationStatus, starred: bool | None = None) -> str:
    verbose_dict: dict[str, str | None] = defaultdict(lambda: None)
    verbose_dict["NotStarted"] = "To Do"
    verbose_dict["InProgress"] = "In Progress"
    verbose_dict["Abandoned"] = "Cancelled"
    verbose_dict["Completed"] = "Completed"
    verbose_dict["InReview"] = "In Review"

    status_value = verbose_dict[status.value]
    if status_value is None:
        raise ValueError("Iteration status unknown")
    if starred is True:
        return status_value + " (*)"
    return status_value


def get_phase_status(status: PhaseStatus) -> str:
    verbose_dict: dict[str, str | None] = defaultdict(lambda: None)
    verbose_dict["NotStarted"] = "Not Started"
    verbose_dict["Draft"] = "In Progress"
    verbose_dict["InReview"] = "In Review"
    verbose_dict["Completed"] = "Completed"

    status_value = verbose_dict[status.value]
    if status_value is None:
        raise ValueError("Phase status unknown")
    return status_value


def get_review_status(status: str) -> str:
    verbose_dict: dict[str, str | None] = defaultdict(lambda: None)
    verbose_dict["Rejected"] = "Rejected"
    verbose_dict["Approved"] = "Approved"
    verbose_dict["Pending"] = "In Review"
    verbose_dict["Cancelled"] = "Cancelled"

    status_value = verbose_dict[status]
    if status_value is None:
        raise ValueError("Review status unknown")
    return status_value



from __future__ import annotations

import logging

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata.base import DatasetSourceOrigin
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.db_metadata import DBMetadata, MetadataDB, TableType

_logger = logging.getLogger(__name__)

from pandas import DataFrame as PandasDataFrame

try:
    from pyspark.sql import DataFrame as SparkDataFrame
except ImportError:
    SparkDataFrame = None
try:
    from h2o.frame import H2OFrame  # pyright: ignore[reportMissingImports]
except:
    H2OFrame = None

TDataFrameType = DataFrameType


class DFResource(Resource):
    """Wrap in-memory DataFrames into a Dataset Resource for metadata extraction.

    This resource is intended for DataFrames that exist only in memory (e.g., Pandas, Spark, or H2O) and are not associated with external file paths or storage locations.
    It enables Vectice to extract schema and optional column-level statistics from these DataFrames, which can be logged as part of a Dataset version.

    Unlike other resource types (e.g., FileResource, S3Resource, BigQueryResource), `DFResource` does not carry path or source metadata—its sole purpose is to wrap raw DataFrames.

    You typically use it when your data is generated on-the-fly, transformed in-memory, or does not have an accessible source path.

    ```python
    from vectice import DFResource, Dataset

    my_df_resource = DFResource(dataframes=df)
    Dataset.clean(name = 'my_dataset', resource = my_df_resource)
    ```

    NOTE: **Vectice does not store your actual dataset. Instead, it stores your dataset's metadata, which is information about your dataset.
    These details are captured by resources tailored to the specific environment in use, such as: local (FileResource), Bigquery, S3, GCS...**
    """

    _origin = DatasetSourceOrigin.LOCAL.value

    def __init__(
        self,
        dataframes: TDataFrameType | list[TDataFrameType],
        capture_schema_only: bool = False,
        columns_description: dict[str, str] | None = None,
    ):
        """Initialize a Dataframe resource resource.

        Parameters:
            dataframes: The dataframes allowing vectice to compute metadata about this resource such as columns schema and optionally statistics. (Support Pandas, Spark, H2O)
            capture_schema_only (Optional): A boolean parameter indicating whether to capture only the schema or both the schema and column statistics of the dataframes.
            columns_description (Optional): A dictionary or path to a csv file to map the column's name to a specific description. Dictionary should follow the format { "column_name": "Description", ... }
        """
        super().__init__(
            paths="",
            dataframes=dataframes,
            capture_schema_only=capture_schema_only,
            columns_description=columns_description,
        )
        first_df = dataframes[0] if isinstance(dataframes, list) else dataframes
        self._type = self._infer_type(first_df)

    def _infer_type(self, df: TDataFrameType) -> TableType:
        if isinstance(df, PandasDataFrame):
            return TableType.PANDAS
        elif SparkDataFrame is not None and isinstance(df, SparkDataFrame):
            return TableType.SPARK
        elif H2OFrame is not None and isinstance(df, H2OFrame):
            return TableType.H2O
        else:
            return TableType.UNKNOWN

    def _fetch_data(
        self,
    ) -> dict:
        return {}

    def _build_metadata(self) -> DBMetadata:
        dbs = []
        if self._dataframes:
            for dataframe in self._dataframes:
                dbs.append(
                    MetadataDB(
                        name="Dataframe",
                        uri="Dataframe",
                        dataframe=dataframe,
                        display_name="Dataframe",
                        capture_schema_only=self.capture_schema_only,
                        columns=[],
                        type=self._type,
                    )
                )

            metadata = DBMetadata(
                size=0,
                origin=self._origin,
                dbs=dbs,
            )
            return metadata
        raise ValueError("A dataframe is required.")



from __future__ import annotations

import logging
import warnings
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, List, TypedDict

from vectice.models.table import Table

_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from numpy.typing import ArrayLike
    from pandas import DataFrame


class ValidationType(Enum):
    """Enumeration of the supported types for validation class."""

    BinaryClassification = "BinaryClassification"


MAP_TYPE = {
    ValidationType.BinaryClassification: ["roc", "cm", "binary_full_suite", "explainability", "feature_importance"]
}


def aggregate_dict(dict1: TestSuiteReturnType, dict2: TestSuiteReturnType) -> TestSuiteReturnType:
    for key, value in dict2["metrics"].items():
        dict1["metrics"][key] = value
    for key, value in dict2["properties"].items():
        dict1["properties"][key] = value

    dict1["tables"].extend(dict2["tables"])
    dict1["attachments"].extend(dict2["attachments"])

    return dict1


class TestSuiteReturnType(TypedDict):
    metrics: dict[str, Any]
    properties: dict[str, Any]
    tables: list[Table]
    attachments: list[str]


class ValidationModel:
    """Represent a validation suite of test to be run."""

    def __init__(
        self,
        asset: str | None,
        training_df: DataFrame,
        testing_df: DataFrame,
        target_column: str,
        predictor: Any,
        predict_proba_test: ArrayLike,
        predict_proba_train: ArrayLike | None = None,
        threshold: float = 0.5,
        type: ValidationType = ValidationType.BinaryClassification,
        tests: List[str | Callable[[Any], TestSuiteReturnType]] = ["roc", "cm"],
    ):
        if asset is None:
            warnings.warn("Asset is None, generated tests will be logged at iteration level.")

        mapped_type = MAP_TYPE[type]
        for test in tests:
            if test not in mapped_type:
                raise ValueError(f"Type {type} is incompatible with test {test}")

        self.asset = asset
        self._training_df = training_df
        self._testing_df = testing_df
        self._predictor = predictor
        self._type = type
        self._predict_proba_train = predict_proba_train
        self._predict_proba_test = predict_proba_test
        self._target = target_column
        self._threshold = threshold
        self._tests = tests

    def execute_test(self) -> TestSuiteReturnType:
        from vectice.models.test_library.binary_classification_test import MAP_TEST

        test_result: TestSuiteReturnType = {
            "metrics": {},
            "properties": {},
            "tables": [],
            "attachments": [],
        }
        for test in self._tests:
            if test in MAP_TEST:
                test_function = MAP_TEST[test]
                if isinstance(test_function, list):
                    for test_func in test_function:
                        temp_return = test_func(
                            self._training_df,
                            self._testing_df,
                            self._target,
                            self._predictor,
                            self._predict_proba_train,
                            self._predict_proba_test,
                            self._threshold,
                        )
                        test_result = aggregate_dict(test_result, temp_return)
                else:
                    temp_return = test_function(
                        self._training_df,
                        self._testing_df,
                        self._target,
                        self._predictor,
                        self._predict_proba_train,
                        self._predict_proba_test,
                        self._threshold,
                    )
                    test_result = aggregate_dict(test_result, temp_return)
            _logger.info(f"Test: {test} successfully run")
            # todo callable
        return test_result



from __future__ import annotations

from typing import Literal

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.gql_model import RETURNS_MODEL_VERSION
from vectice.api.json.issue import IssueOutput
from vectice.api.json.paged_response import PagedResponse
from vectice.utils.api_utils import DEFAULT_NUMBER_OF_ITEMS, PAGINATE_OUTPUT, get_page_input

_RETURNS = """
    id
    name
    severity
    dueDate
    status
    createdDate
    report {
        id
        name
        createdBy {
            name
            email
        }
    }
    owner {
        name
        email
    }
    reviewer {
        name
        email
    }
    __typename
"""

_RETURNS_WITH_MDV = f"""
    {_RETURNS}
    modelVersion {{
        {RETURNS_MODEL_VERSION}
        __typename
    }}
"""

_PAGINATED_ISSUES = PAGINATE_OUTPUT.format(_RETURNS)
_PAGINATED_ISSUES_WITH_MDV = PAGINATE_OUTPUT.format(_RETURNS_WITH_MDV)


class GqlIssueApi(GqlApi):
    def get_issues(
        self, id: str, type: Literal["prj"] | Literal["mdv"], size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[IssueOutput]:
        gql_query = "Findings"
        variable_types = "$type:FindingQueryType,$filters:FindingFiltersInput!,$page:PageInput"
        kw = "type:$type,filters:$filters,page:$page"
        kw_name = "projectId" if type == "prj" else "modelVersionId"
        variables = {
            "type": "PROJECT" if type == "prj" else "MODEL_VERSION",
            "filters": {kw_name: id},
            "page": get_page_input(size=size),
        }
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_PAGINATED_ISSUES if type == "mdv" else _PAGINATED_ISSUES_WITH_MDV,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            return Parser().parse_paged_response(response[gql_query])
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model_version" if type == "mdv" else "project", id)



from __future__ import annotations

from typing import Any

from vectice.autolog.model_library import ModelLibrary


class TechniqueService:

    def _get_model_technique(self, model_object: Any, libary: ModelLibrary):
        if libary is ModelLibrary.STATSMODEL:
            algorithm = str(model_object.model.__class__).split(".")[-1]  # type: ignore[reportAttributeAccessIssue]
        elif libary is ModelLibrary.SKLEARN_PIPELINE:
            try:
                algorithm = str(model_object.steps[-1][-1].__class__).split(".")[-1]  # type: ignore[reportAttributeAccessIssue]
            except Exception:
                algorithm = str(model_object.__class__).split(".")[-1]
        else:
            algorithm = str(model_object.__class__).split(".")[-1]

        return algorithm.replace("'>", "")



from __future__ import annotations

from abc import abstractmethod
from importlib.util import find_spec
from typing import Any, Protocol

from vectice.api.http_error_handlers import VecticeException
from vectice.autolog.asset_services import (
    AutologCatboostService,
    AutologH2oDisplayService,
    AutologH2oEstimatorService,
    AutologH2OFrameService,
    AutologKerasService,
    AutologLightgbmService,
    AutologPandasService,
    AutologPysparkMLService,
    AutologPysparkService,
    AutologPytorchService,
    AutologSklearnService,
    AutologStatsModelWrapperService,
    AutologVecticeAssetService,
    GiskardQATestSetService,
    GiskardRAGReportService,
    GiskardScanReportService,
    GiskardTestSuiteResultService,
    IpywidgetsOutputService,
    ModevaFactSheetService,
    ModevaValidationResult,
    VecticeObjectClasses,
)


class IAutologService(Protocol):
    @abstractmethod
    def get_asset(self) -> dict[str, Any] | None: ...


class AssetFactory:
    @staticmethod
    def get_asset_service(
        key: str,
        asset: Any,
        data: dict,
        phase_name: str,
        custom_metrics: set[str | None] = set(),
        raw_cells_data: list = [],
        capture_schema_only: bool = False,
        prefix: str | None = None,
    ) -> IAutologService:
        is_pandas = find_spec("pandas") is not None
        is_pyspark = find_spec("pyspark") is not None
        is_lgbm = find_spec("lightgbm") is not None
        is_sklearn = find_spec("sklearn") is not None
        is_catboost = find_spec("catboost") is not None
        is_keras = find_spec("keras") is not None
        is_statsmodels = find_spec("statsmodels") is not None
        is_pytorch = find_spec("torch") is not None
        is_modeva = find_spec("modeva") is not None
        is_giskard = find_spec("giskard") is not None
        is_ipywidgets = find_spec("ipywidgets") is not None
        is_h2o = find_spec("h2o") is not None

        if is_giskard:
            from giskard.core.suite import TestSuiteResult  # type: ignore[reportMissingImports]
            from giskard.rag import QATestset  # type: ignore[reportMissingImports]
            from giskard.rag.report import RAGReport  # type: ignore[reportMissingImports]
            from giskard.scanner.report import ScanReport  # type: ignore[reportMissingImports]

            if isinstance(asset, ScanReport):
                return GiskardScanReportService(key, asset, data)
            if isinstance(asset, RAGReport):
                return GiskardRAGReportService(key, asset)
            if isinstance(asset, QATestset):
                return GiskardQATestSetService(key, asset)
            if isinstance(asset, TestSuiteResult):
                return GiskardTestSuiteResultService(key, asset)

        if is_modeva:
            from modeva import FactSheet  # type: ignore[reportMissingImports]
            from modeva.utils.results import ValidationResult  # type: ignore[reportMissingImports]

            if isinstance(asset, FactSheet):
                return ModevaFactSheetService(key, asset, {})
            if isinstance(asset, ValidationResult):
                return ModevaValidationResult(key, asset, {})

        if is_pandas:
            from pandas import DataFrame

            if isinstance(asset, DataFrame):
                return AutologPandasService(key, asset, phase_name, raw_cells_data, capture_schema_only, prefix)

        if is_pyspark:
            # base estimator class, Predictor is for a model with no fit
            from pyspark.ml.base import Model, PredictionModel
            from pyspark.ml.pipeline import PipelineModel
            from pyspark.ml.tuning import CrossValidatorModel

            # pyspark evaluators
            # Pipeline isn't fitted and PipelineModel has a fit
            # clustering models e.g kmeans
            # JavaParams is for pipeline components (might remove/move logic due to overlaps)
            from pyspark.ml.wrapper import JavaModel

            # , Evaluator, JavaParams, Pipeline
            pyspark_service_types = (CrossValidatorModel, PredictionModel, JavaModel, Model, PipelineModel)

            from pyspark.sql import DataFrame as SparkDF
            from pyspark.sql.connect.dataframe import DataFrame as SparkConnectDF

            if isinstance(asset, (SparkDF, SparkConnectDF)):
                return AutologPysparkService(key, asset, phase_name, raw_cells_data, capture_schema_only, prefix)

            if isinstance(asset, pyspark_service_types):
                return AutologPysparkMLService(key, asset, data, custom_metrics, phase_name, prefix)

        if is_lgbm:
            from lightgbm.basic import Booster

            if isinstance(asset, Booster):
                return AutologLightgbmService(key, asset, data, custom_metrics, phase_name, prefix)

        if is_catboost:
            from catboost.core import CatBoost

            if isinstance(asset, CatBoost):
                return AutologCatboostService(key, asset, data, custom_metrics, phase_name, prefix)
        if is_h2o:
            from h2o.display import H2ODisplay  # type: ignore[reportMissingImports]
            from h2o.estimators import H2OEstimator  # type: ignore[reportMissingImports]
            from h2o.frame import H2OFrame  # type: ignore[reportMissingImports]

            # We do not want to accidentally log models or frames as H2ODisplay objects.
            if isinstance(asset, H2ODisplay) and not isinstance(asset, (H2OEstimator, H2OFrame)):
                return AutologH2oDisplayService(key, asset, phase_name, prefix)

            if isinstance(asset, H2OEstimator):
                return AutologH2oEstimatorService(key, asset, data, custom_metrics, phase_name, prefix)

            if isinstance(asset, H2OFrame):
                return AutologH2OFrameService(key, asset, phase_name, raw_cells_data, capture_schema_only, prefix)

        if is_keras:
            from keras.models import Model as KerasModel  # type: ignore[reportMissingImports]

            if isinstance(asset, KerasModel):
                return AutologKerasService(key, asset, data, custom_metrics, phase_name, prefix)

        if is_pytorch:
            from torch.nn import Module

            if isinstance(asset, Module):
                return AutologPytorchService(key, asset, data, custom_metrics, phase_name, prefix)

        if isinstance(asset, VecticeObjectClasses):
            return AutologVecticeAssetService(key, asset)  # type: ignore[reportArgumentType]

        if is_statsmodels:
            from statsmodels.base.wrapper import ResultsWrapper

            if isinstance(asset, ResultsWrapper):
                return AutologStatsModelWrapperService(key, asset, data, custom_metrics, phase_name, prefix)

        if is_ipywidgets:
            from ipywidgets import Output

            if isinstance(asset, Output):
                return IpywidgetsOutputService(key, asset)

        if is_sklearn:
            return AutologSklearnService(key, asset, data, custom_metrics, phase_name, prefix)

        raise VecticeException(f"Asset {asset} of type ({type(asset)!r}) not handled")



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from gql import gql

from vectice.api.http_error_handlers import IDError, MissingReferenceError
from vectice.api.json.paged_response import PagedResponse
from vectice.api.json.project import ProjectCreateInput
from vectice.types.project import ProjectInput
from vectice.utils.api_utils import DEFAULT_NUMBER_OF_ITEMS, PAGINATE_OUTPUT, get_page_input
from vectice.utils.vectice_ids_regex import PROJECT_VID_REG

if TYPE_CHECKING:
    from vectice.api.json.project import ProjectOutput

from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

_RETURNS = """
            vecticeId
            name
            description
            workspace {
                vecticeId
                name
                description
                __typename
            }
            __typename
"""


_RETURNS_LIST = """
            vecticeId
            name
            description
            __typename
"""

_RETURNS_PAGE = PAGINATE_OUTPUT.format(_RETURNS_LIST)


class ProjectApi(GqlApi):
    def get_project(self, project: str, workspace: str | None) -> ProjectOutput:
        search_vectice_id = re.search(PROJECT_VID_REG, project)
        if search_vectice_id:
            gql_query = "getProjectById"
            variable_types = "$projectId:VecticeId!"
            variables = {"projectId": project}
            kw = "projectId:$projectId"
        else:
            if workspace is None:
                raise MissingReferenceError("workspace")
            gql_query = "getProjectByName"
            variable_types = "$name:String!,$workspaceId:VecticeId!"
            variables = {"name": project, "workspaceId": workspace}
            kw = "name:$name,workspaceId:$workspaceId"
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            project_output: ProjectOutput = Parser().parse_item(response[gql_query])
            return project_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "project", project)

    def create_project(self, workspace_id: str, project: ProjectInput) -> ProjectOutput:
        gql_query = "createProject"
        variable_types = "$data:ProjectCreateInput!,$workspaceId:VecticeId!"
        variables = {
            "data": ProjectCreateInput(project),
            "workspaceId": workspace_id,
        }
        kw = "data:$data,workspaceId:$workspaceId"
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=False
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            project_output: ProjectOutput = Parser().parse_item(response[gql_query])
            return project_output
        except TransportQueryError as e:
            status_code, key, message = self._error_handler.get_gql_error_info(e)
            if status_code == 404 and key == "rights_guard":
                raise IDError(message)

            self._error_handler.handle_post_gql_error(e, "workspace", workspace_id)

    def list_projects(self, workspace: str, size: int = DEFAULT_NUMBER_OF_ITEMS) -> PagedResponse[ProjectOutput]:
        gql_query = "getProjectList"
        variable_types = "$page: PageIndexInput!,$workspaceId:VecticeId!"
        variables = {"page": get_page_input(size=size), "workspaceId": workspace}
        kw = "page:$page,workspaceId:$workspaceId"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_PAGE,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            project_output: PagedResponse[ProjectOutput] = Parser().parse_paged_response(response[gql_query])
            return project_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "projects", "list")



from __future__ import annotations

from enum import Enum

from vectice.models.resource.metadata.base import (
    DatasetSourceType,
    DatasetSourceUsage,
    Metadata,
)
from vectice.models.resource.metadata.column_metadata import DBColumn
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.extra_metadata import ExtraMetadata
from vectice.models.resource.metadata.source import Source

TDataFrameType = DataFrameType


class TableType(Enum):
    """Enumeration that defines what the table type is."""

    PANDAS = "PANDAS"
    SPARK = "SPARK"
    H2O = "H2O"
    UNKNOWN = "UNKNOWN"


class DBMetadata(Metadata):
    """Class that describes metadata of dataset that comes from a database."""

    def __init__(
        self,
        dbs: list[MetadataDB],
        size: int | None,
        usage: DatasetSourceUsage | None = None,
        origin: str | None = None,
    ):
        """Initialize a DBMetadata instance.

        Parameters:
            dbs: The list of databases.
            size: The size of the metadata.
            usage: The usage of the metadata.
            origin: The origin of the metadata.
        """
        super().__init__(size=size, type=DatasetSourceType.DB, usage=usage, origin=origin)
        self.dbs = dbs

    def asdict(self) -> dict:
        for db in self.dbs:
            if self._settings is not None:  # pyright: ignore[reportUnnecessaryComparison]
                db.set_settings(self._settings)
        return {
            **super().asdict(),
            "dbs": [db.asdict() for db in self.dbs],
        }


class MetadataDB(Source):
    def __init__(
        self,
        name: str,
        columns: list[DBColumn] | None,
        rows_number: int | None = None,
        size: int | None = None,
        updated_date: str | None = None,
        created_date: str | None = None,
        uri: str | None = None,
        dataframe: TDataFrameType | None = None,
        extra_metadata: list[ExtraMetadata] | None = None,
        display_name: str | None = None,
        capture_schema_only: bool = False,
        type: TableType = TableType.UNKNOWN,
    ):
        """Initialize a MetadataDB instance.

        Parameters:
            name: The name of the table.
            columns: The columns that compose the table.
            rows_number: The number of row of the table.
            size: The size of the table.
            updated_date: The date of last update of the table.
            created_date: The creation date of the table.
            uri: The uri of the table.
            dataframe (Optional): A dataframe allowing vectice to optionally compute more metadata about this resource such as columns stats, size, rows number and column numbers. (Support Pandas and Spark)
            extra_metadata (Optional): Extra metadata to be captured.
            display_name (Optional): Name that will be shown in the Web App.
            capture_schema_only (Optional): A boolean parameter indicating whether to capture only the schema or both the schema and column statistics of the dataframes.
            type (Optional): The table type.
        """
        super().__init__(
            name=name,
            size=size,
            columns=list(columns) if columns is not None else None,
            updated_date=updated_date,
            created_date=created_date,
            uri=uri,
            dataframe=dataframe,
            extra_metadata=extra_metadata,
            display_name=display_name,
            capture_schema_only=capture_schema_only,
        )
        self.rows_number = rows_number
        self.type = type

    def asdict(self) -> dict:
        return {
            "rowsNumber": self.rows_number,
            "columnsNumber": len(self.columns) if self.columns is not None else None,
            "type": self.type.value,
            **super().asdict(),
            "tablename": self.display_name,
        }



from __future__ import annotations

import logging
from importlib.util import find_spec

_logger = logging.getLogger(__name__)

is_plotly = find_spec("plotly") is not None
is_matplotlib = find_spec("matplotlib") is not None


class PlotlyPatcher:
    """Handles monkey patching of Plotly's write_image method."""

    def __init__(self, patched_graphs: list[str]):
        self._original_function = None
        self._is_patched = False
        self._patched_graphs = patched_graphs

    def patch(self):
        """Apply the patch to Plotly's write_image method."""
        if not is_plotly or self._is_patched:
            return

        import plotly.graph_objects as go

        original_method = go.Figure.write_image
        self._original_function = original_method
        patched_graphs = self._patched_graphs

        def patched_write_image(self, *args, **kwargs):  # pyright: ignore[reportMissingParameterType]
            try:
                graph_path = args[0] if args else kwargs.get("file")
                if graph_path:
                    patched_graphs.append(graph_path)
            except Exception:
                pass
            return original_method(self, *args, **kwargs)

        go.Figure.write_image = patched_write_image
        self._is_patched = True

    def unpatch(self):
        """Restore the original write_image method."""
        if not is_plotly or not self._is_patched or not self._original_function:
            return

        import plotly.graph_objects as go

        go.Figure.write_image = self._original_function
        self._is_patched = False


class MatplotlibPatcher:
    """Handles monkey patching of Matplotlib's savefig function."""

    def __init__(self, patched_graphs: list[str]):
        self._original_function = None
        self._is_patched = False
        self._patched_graphs = patched_graphs

    def patch(self):
        """Apply the patch to matplotlib's savefig function."""
        import matplotlib.pyplot as plt

        if not is_matplotlib or self._is_patched:
            return

        original_function = plt.savefig
        self._original_function = original_function

        def patched_savefig(fname, *args, **kwargs):  # pyright: ignore[reportMissingParameterType]
            """A patched version of matplotlib's savefig function that tracks the file path."""
            try:
                if fname:
                    self._patched_graphs.append(fname)
            except Exception:
                pass
            return original_function(fname, *args, **kwargs)

        plt.savefig = patched_savefig
        self._is_patched = True

    def unpatch(self):
        """Restore the original savefig function."""
        import matplotlib.pyplot as plt

        if not is_matplotlib or not self._is_patched or not self._original_function:
            return

        plt.savefig = self._original_function
        self._is_patched = False


class GraphPatchService:
    """Manages all graph library patches."""

    def __init__(self):
        self._patched_graphs: list[str] = []
        self.plotly_patcher = PlotlyPatcher(self._patched_graphs)
        self.matplotlib_patcher = MatplotlibPatcher(self._patched_graphs)

    @property
    def graphs(self) -> list[str]:
        """Get the list of captured graph file paths."""
        return self._patched_graphs.copy()

    def clear_graphs(self):
        """Clear the collected graphs list."""
        self._patched_graphs.clear()

    def patch_all(self):
        """Apply all patches."""
        self.clear_graphs()

        self.plotly_patcher.patch()
        self.matplotlib_patcher.patch()

    def unpatch_all(self):
        """Remove all patches."""
        self.plotly_patcher.unpatch()
        self.matplotlib_patcher.unpatch()



from __future__ import annotations

from datetime import datetime
from typing import Any

from vectice.api._utils import read_nodejs_date
from vectice.api.json.json_type import TJSON
from vectice.utils.logging_utils import get_review_status


class ReviewOutput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def message(self) -> str | None:
        value = self["message"]
        return str(value) if value else None

    @property
    def feedback(self) -> str | None:
        value = self["feedback"]
        return str(value) if value else None

    @property
    def created_date(self) -> datetime | None:
        if self["createdDate"] is not None:
            return read_nodejs_date(str(self["createdDate"]))
        return None

    @property
    def status(self) -> str:
        return get_review_status(self["status"])

    @property
    def creator(self) -> dict[str, str]:
        return {"name": self["createdBy"]["name"], "email": self["createdBy"]["email"]}

    @property
    def assignee(self) -> dict[str, str]:
        return {"name": self["reviewer"]["name"], "email": self["reviewer"]["email"]}



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from vectice.api.json.issue import IssueOutput
from vectice.models.representation.report_representation import ReportRepresentation
from vectice.utils.common_utils import (
    repr_class,
)

if TYPE_CHECKING:
    from vectice.api.client import Client


class IssueRepresentation:
    """Represents the metadata of a Vectice issue.

    An Issue Representation shows information about a specific issue from the Vectice app.
    It makes it easier to get and read this information through the API.

    Attributes:
        id (int): The unique identifier of the issue.
        name (str): The name of the issue.
        status (str): The status of the issue (Cancelled, Completed, InProgress, NotStarted, or RemediationPending).
        severity (str | None): The severity of the issue (Low, Medium, or High).
        due_date (datetime | None): The due date of the issue.
        model_version (ModelVersionRepresentation | None): The model version related to this finding.
        created_date (datetime | None): The date when the issue was created.
        report (ReportRepresentation | None): The report related to this finding.
        owner (Dict[str, str]): Owner of the issue.
        assignee (Dict[str, str]): Assignee of the issue.
    """

    def __init__(self, output: IssueOutput, client: Client):
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation

        self.id = output.id
        self.name = output.name
        self.status = output.status
        self.severity = output.severity
        self.due_date = output.due_date
        self.model_version = (
            ModelVersionRepresentation(output.model_version, client) if output.model_version is not None else None
        )
        self.created_date = output.created_date
        self.report = ReportRepresentation(output.report, client) if output.report is not None else None
        self.owner = output.owner
        self.assignee = output.assignee

    def __repr__(self):
        return repr_class(self)

    def asdict(self) -> Dict[str, Any]:
        """Transform the IssueRepresentation into a organised dictionary.

        Returns:
            The object represented as a dictionary
        """
        return {
            "id": self.id,
            "name": self.name,
            "status": self.status,
            "severity": self.severity,
            "due_date": self.due_date,
            "model_version": (self.model_version.asdict() if self.model_version else None),
            "created_date": self.created_date,
            "report": (self.report.asdict() if self.report else None),
            "owner": self.owner,
            "assignee": self.assignee,
        }



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from vectice.api._utils import read_nodejs_date
from vectice.api.json.json_type import TJSON

if TYPE_CHECKING:
    from datetime import datetime

    from vectice.api.json import ProjectOutput


class ModelOutput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        from vectice.api.json import ProjectOutput

        super().__init__(*args, **kwargs)
        if "project" in self:
            self._project: ProjectOutput = ProjectOutput(**self["project"])

    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def deleted_date(self) -> datetime | None:
        return read_nodejs_date(str(self["deletedDate"]))

    @property
    def version(self) -> int:
        return int(self["version"])

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def project(self) -> ProjectOutput:
        from vectice.api.json import ProjectOutput

        return ProjectOutput(**self["project"])

    @project.setter
    def project(self, value: ProjectOutput) -> None:
        self._project = value

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def description(self) -> str:
        return str(self["description"])

    @property
    def author_id(self) -> int:
        return int(self["authorId"])

    @property
    def project_id(self) -> int:
        return int(self["projectId"])



from __future__ import annotations

import os
from textwrap import dedent
from typing import TYPE_CHECKING

from PIL import Image

from vectice.models.representation.dataset_representation import DatasetRepresentation
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_representation import ModelRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation

if TYPE_CHECKING:
    from logging import Logger

    from vectice.api import Client
    from vectice.api.json.dataset_register import DatasetRegisterOutput
    from vectice.api.json.model_register import ModelRegisterOutput
    from vectice.models.dataset import Dataset
    from vectice.models.iteration import Iteration
    from vectice.models.model import Model
    from vectice.models.table import Table


def get_last_user_and_default_workspace(client: Client) -> tuple[str, str | None]:
    asset = client.get_user_and_default_workspace()
    workspace_id = str(asset["defaultWorkspace"]["vecticeId"]) if asset["defaultWorkspace"] else None
    return asset["user"]["name"], workspace_id


def connection_logging(_logger: Logger, user_name: str, host: str, workspace_id: str | None):
    from vectice.utils.logging_utils import CONNECTION_LOGGING

    if workspace_id:
        logging_output = f"For quick access to your default workspace in the Vectice web app, visit:\n{host}/browse/workspace/{workspace_id}"
        _logger.info(CONNECTION_LOGGING.format(user=user_name, logging_output=logging_output))
        return
    logging_output = f"For quick access to the list of workspaces in the Vectice web app, visit:\n{host}/workspaces"
    _logger.info(CONNECTION_LOGGING.format(user=user_name, logging_output=logging_output))


def _get_iteration_log(iteration: Iteration, section: str | None = None, existing_asset: bool = False):
    iteration_id = iteration.id
    url = iteration._client.auth.api_base_url  # pyright: ignore[reportPrivateUsage] OK to ignore deprecated code
    hyper_link = f"{url}/browse/iteration/{iteration_id}"
    assignment_type = "reassigned to" if existing_asset else "in"
    to_section = f", {assignment_type} section: {section!r}." if section else "."
    return f"to iteration: {iteration.name!r}{to_section}", f"Link to iteration: {hyper_link}"


def _get_image_or_file_artifact_reference(filename: str) -> str:
    _, ext = os.path.splitext(filename)
    pillow_extensions = {exten for exten in Image.registered_extensions()}
    return "image" if (ext in pillow_extensions and ext != ".pdf") else "file"


def comment_or_image_logging(
    iteration: Iteration,
    _logger: Logger,
    section: str | None,
    filename: str | None = None,
):
    artifact_reference = "comment"
    if filename:
        asset_type_log = _get_image_or_file_artifact_reference(filename)
        artifact_reference = f"{asset_type_log}: {filename!r}"

    assignment_log, iteration_link_log = _get_iteration_log(iteration, section)
    logging_output = dedent(
        f"""
            Added {artifact_reference} {assignment_log}
            {iteration_link_log}
        """
    ).lstrip()

    _logger.info(logging_output)


def table_logging(
    iteration: Iteration,
    data: Table,
    _logger: Logger,
    section: str | None,
):
    name = data.name

    assignment_log, iteration_link_log = _get_iteration_log(iteration, section)
    logging_output = dedent(
        f"""
            Table {name} added {assignment_log}
            {iteration_link_log}
        """
    ).lstrip()

    _logger.info(logging_output)


def _get_full_asset_log(
    asset_log: str,
    iteration: Iteration,
    section: str | None,
    attachments: list[str] | None,
    existing_asset: bool = False,
):
    attachments_output = ", ".join([attach for attach in attachments]) if attachments else "None"
    assignment_log, iteration_link_log = _get_iteration_log(iteration, section, existing_asset)
    return dedent(
        f"""
            {asset_log} {assignment_log}
            Attachments: {attachments_output}
            {iteration_link_log}
        """
    ).lstrip()


def _get_register_dataset_log(data: DatasetRegisterOutput, value: Dataset) -> str:
    dsv_name = data.dataset_version.name
    if data.use_existing_version:
        return f"Existing dataset: {value.name!r} and version: {dsv_name!r} already linked"
    if data.use_existing_dataset:
        return f"New version: {dsv_name!r} of dataset: {value.name!r} added"

    return f"New dataset: {value.name!r} version: {dsv_name!r} added"


def register_dataset_logging(
    iteration: Iteration,
    data: DatasetRegisterOutput,
    value: Dataset,
    attachments: list[str] | None,
    _logger: Logger,
    section: str | None = None,
):
    asset_log = _get_register_dataset_log(data, value)
    extisting_asset: bool = True if data.use_existing_version else False
    logging_output = _get_full_asset_log(asset_log, iteration, section, attachments, extisting_asset)
    _logger.info(logging_output)


def register_model_logging(
    iteration: Iteration,
    data: ModelRegisterOutput,
    value: Model,
    attachments: list[str] | None,
    success_pickle: bool,
    _logger: Logger,
    section: str | None = None,
):
    mv_name = data.model_version.name
    asset_log = (
        f"New version: {mv_name!r} of model: {value.name!r} added"
        if data.use_existing_model
        else f"New model: {value.name!r} version: {mv_name!r} added"
    )
    logging_output = _get_full_asset_log(asset_log, iteration, section, attachments)

    if not success_pickle:
        _logger.warning(
            "The predictor cannot be serialized. Ensure it is picklable if you intend to log it as an attachment to the model."
        )
    _logger.info(logging_output)


def _get_asset_parent_name(value: ModelVersionRepresentation | DatasetVersionRepresentation):
    return (
        value.model_representation.name
        if isinstance(value, ModelVersionRepresentation)
        else value.dataset_representation.name
    )


def assign_asset_version_logging(
    iteration: Iteration,
    value: ModelRepresentation | ModelVersionRepresentation | DatasetRepresentation | DatasetVersionRepresentation,
    _logger: Logger,
    already_assigned: bool = False,
    section: str | None = None,
):
    assignment_log, iteration_link_log = _get_iteration_log(iteration=iteration, section=section, existing_asset=True)
    asset_info = "model" if isinstance(value, (ModelVersionRepresentation, ModelRepresentation)) else "dataset"
    link = "already linked" if already_assigned else "linked"

    is_parent = isinstance(value, (ModelRepresentation, DatasetRepresentation))
    asset_id = (
        value._last_version.id  # pyright: ignore [reportPrivateUsage]
        if (is_parent and value._last_version is not None)  # pyright: ignore [reportPrivateUsage]
        else value.id
    )
    asset_name = value.name if is_parent else _get_asset_parent_name(value)

    logging_output = dedent(
        f"""
            Existing {asset_info} version: {asset_id!r} of {asset_info}: {asset_name!r} {link} {assignment_log}
            {iteration_link_log}
        """
    ).lstrip()
    _logger.info(logging_output)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.model_representation import ModelRepresentationOutput
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput, ModelVersionUpdateInput
from vectice.api.json.paged_response import PagedResponse
from vectice.utils.api_utils import DEFAULT_NUMBER_OF_ITEMS, PAGINATE_OUTPUT, get_page_input

if TYPE_CHECKING:
    from vectice.api.json.iteration import IterationContextInput
    from vectice.api.json.model_register import ModelRegisterInput, ModelRegisterOutput


_RETURNS_MODEL_VERSION_APPROVAL_HISTORY = """
    toValidateDate
    inValidationDate
    validatedDate
    approvedDate
    rejectedDate
"""

_RETURNS = """
            modelVersion {
                vecticeId
                name
                version
                description
                algorithmName
                framework
                modelId
                model {
                    name
                    projectId
                }
                origin {
                   id
                }
            }
            useExistingModel
            __typename
            """

_RETURNS_MODEL = """
            vecticeId
            createdDate
            updatedDate
            name
            description
            type
            versionCount
            versionStats {
                experimentationCount
                productionCount
                stagingCount
                retiredCount
                discardedCount
            }
            lastVersion {
                vecticeId
                createdDate
                updatedDate
                name
                description
                status
                algorithmName
                framework
                __typename
            }
            project {
                vecticeId
            }
            __typename
"""

_BASE_MODEL_VERSION = """
            vecticeId
            name
            isStarred
            description
            status
            risk
            approval
            algorithmName
            framework           
            inventoryReference 
            createdBy {
                name
                email
            }
            properties {
                key
                value
            }
            metrics {
                key
                value
            }
            origins {
                phase {
                    vecticeId
                    __typename
                }
                iteration {
                    vecticeId
                    __typename
                }
                __typename
            }
            __typename
"""
RETURNS_MODEL_VERSION = f"""
            {_BASE_MODEL_VERSION}
            origins {{
                phase {{
                    vecticeId
                    __typename
                }}
                iteration {{
                    vecticeId
                    __typename
                }}
                __typename
            }}
            model {{
                name
                project {{
                    vecticeId
                }}
                vecticeId
                type
                description
                versionCount
                versionStats {{
                    experimentationCount
                    productionCount
                    stagingCount
                    retiredCount
                    discardedCount
                }}
                lastVersion {{
                    vecticeId
                    __typename
                }}
            }}
"""

_RETURNS_MODEL_UPDATE = """
            vecticeId
            __typename
"""
_RETURNS_PAGINATED_MODEL_LIST = PAGINATE_OUTPUT.format(_RETURNS_MODEL)
_RETURNS_PAGINATED_VERSION_LIST = PAGINATE_OUTPUT.format(_BASE_MODEL_VERSION)


class GqlModelApi(GqlApi):
    def get_model(self, id: str) -> ModelRepresentationOutput:
        variable_types = "$modelId:VecticeId!"
        kw = "modelId:$modelId"
        variables = {"modelId": id}
        query = GqlApi.build_query(
            gql_query="getModelById",
            variable_types=variable_types,
            returns=_RETURNS_MODEL,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            model_output: ModelRepresentationOutput = Parser().parse_item(response["getModelById"])
            return model_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model", id)

    def get_model_list(
        self, project_id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[ModelRepresentationOutput]:
        variable_types = "$projectId:VecticeId!,$page:PageInput"
        kw = "projectId:$projectId,page:$page"
        variables = {
            "projectId": project_id,
            "page": get_page_input(size=size),
        }
        query = GqlApi.build_query(
            gql_query="getModelList",
            variable_types=variable_types,
            returns=_RETURNS_PAGINATED_MODEL_LIST,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            model_output: PagedResponse[ModelRepresentationOutput] = Parser().parse_paged_response(
                response["getModelList"]
            )
            return model_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "project", project_id)

    def get_model_version_list(self, id: str, size: int) -> PagedResponse[ModelVersionRepresentationOutput]:
        gql_query = "getModelVersionsList"
        variable_types = "$modelId:VecticeId!,$page:PageInput"
        kw = "modelId:$modelId,page:$page"
        variables = {
            "modelId": id,
            "page": get_page_input(size=size),
        }
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_PAGINATED_VERSION_LIST,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            model_output: PagedResponse[ModelVersionRepresentationOutput] = Parser().parse_paged_response(
                response[gql_query]
            )
            return model_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model_version", id)

    def get_model_version(self, id: str) -> ModelVersionRepresentationOutput:
        variable_types = "$modelVersionId:VecticeId!"
        kw = "modelVersionId:$modelVersionId"
        variables = {"modelVersionId": id}
        query = GqlApi.build_query(
            gql_query="getModelVersion",
            variable_types=variable_types,
            returns=RETURNS_MODEL_VERSION,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            model_output: ModelVersionRepresentationOutput = Parser().parse_item(response["getModelVersion"])
            return model_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model_version", id)

    def register_model(
        self,
        data: ModelRegisterInput,
        iteration_context: IterationContextInput,
    ) -> ModelRegisterOutput:
        variables: dict[str, Any] = {"iterationContext": iteration_context, "data": data}
        kw = "iterationContext:$iterationContext,data:$data"
        variable_types = "$iterationContext:IterationContextInput!,$data:ModelRegisterInput!"

        query_name = "registerModel"
        query = GqlApi.build_query(
            gql_query=query_name,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            model_output: ModelRegisterOutput = Parser().parse_item(response[query_name])
            return model_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model", "register model")

    def update_model(self, model_version_id: str, model: ModelVersionUpdateInput):
        variable_types = "$modelVersionId:VecticeId!,$data:ModelVersionUpdateInput!"
        kw = "modelVersionId:$modelVersionId,data:$data"
        variables = {"modelVersionId": model_version_id, "data": model}
        query = GqlApi.build_query(
            gql_query="updateModelVersion",
            variable_types=variable_types,
            returns=_RETURNS_MODEL_UPDATE,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            self.execute(query_built, variables)
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model", "put")

    def get_model_version_approval_history(self, id: str) -> dict[str, str | None]:
        gql_query = "ModelVersionApprovalStatusDates"
        variable_types = "$modelVersionId:VecticeId!"
        kw = "modelVersionId:$modelVersionId"
        variables = {"modelVersionId": id}
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_MODEL_VERSION_APPROVAL_HISTORY,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            return response[gql_query]
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model_version", id)



from __future__ import annotations

import json
import logging
import re
from base64 import b64decode
from datetime import datetime

import requests
import urllib3

from vectice.__version__ import __vectice_version__, __version__

DEFAULT_API_ENDPOINT = "https://app.vectice.com"


def default_http_headers() -> dict[str, str]:
    return {"Vectice-SDK-Version": __version__, "Vectice-Version": __vectice_version__}


def format_url(url: str) -> str:
    """Add HTTPS protocol if missing and remove trailing slash.

    Parameters:
        url: The URL to format.

    Returns:
        The formatted URL.
    """
    url = url.rstrip("/")
    if not re.match("(?:http|https|ftp)://", url):
        return "https://{}".format(url)
    return url


_logger = logging.getLogger(__name__)


class Auth:  # nosec B107
    def __init__(
        self,
        api_endpoint: str,
        api_token: str,
    ):
        self._default_request_headers: dict[str, str] = default_http_headers()
        self._API_TOKEN = api_token
        self._API_BASE_URL = format_url(api_endpoint)
        self.verify_certificate = False
        self._jwt: str | None = None
        self._jwt_expiration: int | None = None
        urllib3.disable_warnings()
        self._refresh_token()

    @property
    def _token(self) -> str | None:
        if self._jwt_expiration is None:
            return None
        # Refresh token 1 min before expiration
        if datetime.now().timestamp() >= self._jwt_expiration - 60:
            self._refresh_token()
        return self._jwt

    @_token.setter
    def _token(self, jwt: str) -> None:
        self._jwt = jwt
        self._jwt_expiration = self._get_jwt_expiration(jwt)
        self._default_request_headers["Authorization"] = "Bearer " + jwt

    def _refresh_token(self) -> None:
        _logger.debug("Vectice: Refreshing token... ")
        url = self._API_BASE_URL + "/metadata/authenticate"
        data = '{"apiKey":  "%s" }' % self._API_TOKEN
        headers = {**self._default_request_headers, "Content-Type": "application/json"}
        try:
            response = requests.post(url=url, data=data, verify=self.verify_certificate, headers=headers)  # noqa: S113
            if response.status_code == 200:
                self._token = response.json()["token"]
                _logger.debug("Vectice successfully connected.")
            elif response.status_code == 401:
                raise ConnectionRefusedError("The API token provided is not valid.")
            else:
                raise ValueError(response.text)
        except requests.ConnectionError:
            raise ConnectionError(
                f"Host {self._API_BASE_URL} is not reachable, if you are running your own instance of vectice please indicate it with the host parameter of connect function. You can find more information here : https://docs.vectice.com"
            ) from None

    @staticmethod
    def _get_jwt_expiration(jwt: str) -> int:
        jwt_payload = jwt.split(".")[1]
        jwt_payload_with_padding = f"{jwt_payload}{'=' * (4 - len(jwt_payload) % 4)}"
        return int(json.loads(b64decode(jwt_payload_with_padding))["exp"])

    @property
    def api_base_url(self) -> str:
        return self._API_BASE_URL

    @property
    def http_headers(self) -> dict[str, str]:
        # ensure token is up to date
        self._token  # noqa: B018
        return self._default_request_headers



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from vectice.api.json.json_type import TJSON

if TYPE_CHECKING:
    from vectice.api.json.dataset_version import DatasetVersionOutput
    from vectice.models.property import Property


class DatasetRegisterOutput(TJSON):
    @property
    def dataset_version(self) -> DatasetVersionOutput:
        from vectice.api.json.dataset_version import DatasetVersionOutput

        return DatasetVersionOutput(**self["datasetVersion"])

    @property
    def use_existing_version(self) -> bool:
        return bool(self["useExistingVersion"])

    @property
    def use_existing_dataset(self) -> bool:
        return bool(self["useExistingDataset"])

    # TODO: complete the jobrun property
    @property
    def job_run(self) -> str:
        return str(self["jobRun"])


class DatasetRegisterInput(TJSON):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def data_sources(self) -> list[dict[str, Any]]:
        return list(self["datasetSources"])

    @property
    def dataset_inputs(self) -> list[str] | None:
        return list(self["datasetInputs"]) if self["datasetInputs"] else None

    @property
    def model_inputs(self) -> list[str] | None:
        return list(self["modelInputs"]) if self["modelInputs"] else None

    @property
    def properties(self) -> Property:
        from vectice.models.property import Property

        return Property(**self["properties"])



from __future__ import annotations

from typing import Literal

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.gql_dataset import RETURNS_DATASET_VERSION
from vectice.api.gql_model import RETURNS_MODEL_VERSION
from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput

_RETURNS_INPUTS = f"""
    vecticeId
    inputs {{
        {RETURNS_DATASET_VERSION}
        __typename
    }}
    __typename
"""

_RETURNS_CHILDREN = f"""
    vecticeId
    children {{
        modelVersions {{
            {RETURNS_MODEL_VERSION}
            __typename
        }}
        datasetVersions {{
            {RETURNS_DATASET_VERSION}
            __typename
        }}
        __typename
    }}
    __typename
"""


class GqlLineageApi(GqlApi):
    def get_lineage_inputs(
        self, id: str, type: Literal["dtv"] | Literal["mdv"]
    ) -> list[DatasetVersionRepresentationOutput]:
        gql_query = "getModelVersion" if type == "mdv" else "getDatasetVersion"
        kw_name = "modelVersionId" if type == "mdv" else "datasetVersionId"
        variable_types = f"${kw_name}:VecticeId!"
        kw = f"{kw_name}:${kw_name}"
        variables = {kw_name: id}
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_INPUTS,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            return Parser().parse_list(response[gql_query]["inputs"])
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model_version" if type == "mdv" else "dataset_version", id)

    def get_lineage_children(self, id: str, type: Literal["dtv"] | Literal["mdv"]) -> tuple[
        list[ModelVersionRepresentationOutput],
        list[DatasetVersionRepresentationOutput],
    ]:
        gql_query = "getModelVersion" if type == "mdv" else "getDatasetVersion"
        kw_name = "modelVersionId" if type == "mdv" else "datasetVersionId"
        variable_types = f"${kw_name}:VecticeId!"
        kw = f"{kw_name}:${kw_name}"
        variables = {kw_name: id}
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_CHILDREN,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            return Parser().parse_list(response[gql_query]["children"]["modelVersions"]), Parser().parse_list(
                response[gql_query]["children"]["datasetVersions"]
            )
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model_version" if type == "mdv" else "dataset_version", id)



from __future__ import annotations

import logging
from typing import Any

from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.property_service import PropertyService
from vectice.autolog.asset_services.service_types.vectice_types import VecticeType
from vectice.autolog.asset_services.technique_service import TechniqueService
from vectice.autolog.model_library import ModelLibrary
from vectice.models.model import Model
from vectice.utils.common_utils import flatten_dict, get_asset_name, temp_directory

_logger = logging.getLogger(__name__)

H2O_TEMP_DIR = "h2o"


class AutologH2oEstimatorService(MetricService, PropertyService, TechniqueService):
    def __init__(
        self,
        key: str,
        asset: Any,
        data: dict,
        custom_metrics_data: set[str | None],
        phase_name: str,
        prefix: str | None = None,
    ):
        self._asset = asset
        self._key = key
        self._temp_dir = temp_directory(H2O_TEMP_DIR)
        self._model_name = get_asset_name(self._key, phase_name, prefix)

        super().__init__(cell_data=data, custom_metrics=custom_metrics_data)

    def get_asset(self):
        temp_files = []
        library = ModelLibrary.H2O

        try:
            params = flatten_dict(
                {
                    str(key): value
                    for key, value in self._asset.get_params().items()  # pyright: ignore[reportAttributeAccessIssue]
                    if value is not None and bool(str(value))
                }
            )
            model = Model(
                library=library.value,
                technique=self._get_model_technique(self._asset, ModelLibrary.H2O),
                metrics=self._get_model_metrics(self._cell_data),
                properties=params,
                name=self._model_name,
                predictor=self._asset,
                attachments=temp_files,
            )
            return {
                "variable": self._key,
                "model": model,
                "asset_type": VecticeType.MODEL,
            }
        except Exception as e:
            _logger.debug(f"Failed to get asset from {self._asset.__class__.__name__}: {e!s}.")



from __future__ import annotations

from typing import Any

from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.property_service import PropertyService
from vectice.autolog.asset_services.service_types.modeva_types import ModevaType
from vectice.autolog.asset_services.technique_service import TechniqueService


class ModevaFactSheetService(PropertyService, MetricService, TechniqueService):
    def __init__(self, key: str, asset: Any, data: dict):
        self._asset = asset
        self._key = key

        super().__init__(cell_data=data)

    def _get_model_info(self, model: Any):
        # estimator e.g sklearn / xgboost etc
        library = model.name.replace("'>", "")
        # Modeva wrapper
        technique = str(model.__class__).split(".")[-1].replace("'>", "")
        _, params = self._get_sklearn_or_xgboost_or_lgbm_info(model)
        return {
            "variable": self._key,
            "model": model,
            "library": library,
            "technique": technique,
            # "metrics": self._get_model_metrics(self._cell_data),
            "properties": params,
        }

    def get_asset(self):
        try:
            assets = {
                "variable": self._key,
                "dataset": None,
                "models": None,
                "validation_results": None,
                "asset_type": ModevaType.FACTSHEET,
            }
            # returns dataset / ds.data -> pd.DF
            dataset = self._asset.get_dataset().data
            assets["dataset"] = dataset

            # return estimator
            model = self._asset._LocalFactSheet__model
            if hasattr(self._asset, "_validation_test_results"):
                assets["validation_results"] = self._asset._validation_test_results
            if model:
                model_asset = self._get_model_info(model)
                assets["models"] = [model_asset]  # pyright: ignore[reportArgumentType]
                return assets
            if model is None:
                models = self._asset._LocalFactSheet__models
                model_assets = [self._get_model_info(model) for model in models]
                assets["models"] = model_assets  # pyright: ignore[reportArgumentType]
                return assets
        except Exception:
            pass



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyspark import pandas as ps
    from pyspark.pandas.frame import DataFrame as PysparkPandasDF


class TimeSeries:
    def __init__(self):
        pass

    def __len__(self) -> int:
        return 0

    def max(self) -> datetime:
        return datetime.now()

    def min(self) -> datetime:
        return datetime.now()

    def to_frame(self) -> PysparkPandasDF:
        return PysparkPandasDF()


class Series:
    def __init__(self):
        pass

    def isnull(self) -> Series:
        return Series()

    def sum(self) -> int:
        return 0

    @property
    def index(self) -> ps.Index:
        return ps.Index([0])

    def astype(self, type: str) -> TimeSeries:
        return TimeSeries()



import os
import sys
from ast import parse

from setuptools import setup

if sys.version_info < (3, 9):
    print("vectice requires Python 3 version >= 3.9", file=sys.stderr)
    sys.exit(1)

package_root = os.path.abspath(os.path.dirname(__file__))

version_requires = ">=3.9.0"

with open(os.path.join("src", "vectice", "__version__.py")) as f:
    version = parse(next(line for line in f if line.startswith("__version__"))).body[0].value.s

with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()

setup(
    name="vectice",
    version=version,
    description="Vectice Python library",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Vectice Inc.",
    author_email="sdk@vectice.com",
    url="https://www.vectice.com",
    package_data={"vectice": ["py.typed"]},
    license="Apache License 2.0",
    keywords=["Vectice", "Client", "API", "Adapter"],
    platforms=["Linux", "MacOS X", "Windows"],
    python_requires=version_requires,
    install_requires=[
        "types-python-dateutil",
        "python-dotenv>=0.11.0",
        "requests<=2.32.4",
        "rich",
        "urllib3<=2.5.0",
        "gql[requests]>=4.0.0",
        "GitPython",
        "matplotlib",
        "packaging",
        "Pillow",
        "pandas",
        "typing-extensions>=4.5.0",  # Prior to 4.6.2 because of colab issues with tensorflow 2.13.0
        "dataclasses-json==0.5.8",
        "IPython",
        "ipynbname",
    ],
    extras_require={
        "dev": [
            "black==24.2.0",
            "gitpython",
            "pyright==1.1.360",
            "ruff",
            "types-requests",
            "types-urllib3",
            "types-mock",
            "mypy_boto3_s3==1.34.65",
            "pandas-stubs",
        ],
        "doc": [
            "black==24.2.0",
            "markdown-callouts>=0.2",
            "markdown-exec>=1.2",
            "mkdocs",
            "mkdocs-material>=7.3",
            "mkdocs-redirects>=1.2",
            "mkdocs-section-index>=0.3",
            "mkdocstrings[python]>=0.18",
        ],
        "test": [
            "docker==7.1.0",
            "docker-compose",
            "dslr[psycopg2-binary]",
            "mock>=1.0.1",
            "numpy",
            "pytest==8.2.0",
            "pytest-randomly",
            "coverage",
            "pytest-cov",
            "pydrive2",
            "scikit-learn<=1.6.1",
            "testcontainers",
            "db-dtypes>=1.1.1",
            "pyspark",
            "mlflow<=2.22.1",
            "Cython>=3.0",
            "pyyaml==5.3.1",
            "pyarrow>=15.0.0; python_version >= '3.12'",
            "IPython",
            "matplotlib",
            "papermill",
            "tenacity==8.3.0",
            "plotly",
            "ipynbname",
            "nbformat",
            "lightgbm",
            "catboost",
            "tensorflow>=2.15.1",
            "keras",
            "snowflake-snowpark-python",
            "seaborn",
            "statsmodels",
            "xgboost",
            "kaleido",
            "torch",
            "ipywidgets",
            "polars",
            "wandb",
        ],
        "gcs": ["google-cloud-storage>=1.17.0", "google-cloud-bigquery", "protobuf"],
        "s3": ["boto3"],
        "validation": ["shap>=0.44.1", "scipy<=1.12"],
    },
    classifiers=[
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Typing :: Typed",
    ],
)



from __future__ import annotations

from vectice.api.json.compatibility import CompatibilityOutput
from vectice.api.rest_api import RestApi


class CompatibilityApi(RestApi):
    def check_version(self) -> CompatibilityOutput:
        response = self.get("/metadata/compatibility")
        if "message" not in response.keys():
            return CompatibilityOutput(message="", status="OK")
        else:
            return CompatibilityOutput(**response)



from __future__ import annotations

from typing import Any

from vectice.autolog.asset_services.service_types.giskard_types import GiskardType


class GiskardQATestSetService:
    def __init__(self, key: str, asset: Any):
        self._asset = asset
        self._key = key

    def get_asset(self):
        from vectice import Table

        assets = {
            "variable": self._key,
            "table": None,
            "dataframe": None,
            "csv": None,
            "asset_type": GiskardType.QATESTSET,
        }
        try:
            testset_df = self._asset.to_pandas()
            assets["table"] = Table(testset_df, "QATestset")  # type: ignore[reportArgumentType]
            assets["dataframe"] = testset_df
            filename = "qatestset.csv"
            testset_df.to_csv(filename)
            assets["csv"] = filename
            return assets
        except Exception:
            pass



from vectice.autolog.autolog import (
    add_custom_metric_function,
    cell,
    config,
    get_connection,
    iteration,
    notebook,
    generate_doc,
    phase_config,
    vectice_config,
)

__all__ = [
    "cell",
    "notebook",
    "generate_doc",
    "config",
    "vectice_config",
    "phase_config",
    "get_connection",
    "iteration",
    "add_custom_metric_function",
]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum


@dataclass
class AdditionalInfo:
    """Defines additional information for a model about his context that a user would like to display in vectice.

    Parameters:
        url: The URL to identify the property.
        run_id: The run ID of the property.
        framework: The framework used for the property.
        extra_info: Additional information related to the model run.
    """

    url: str | None = None
    run: str | None = None
    framework: Framework | None = None
    extra_info: ExtraInfo | list[ExtraInfo] | None = None

    def __post_init__(self):
        if self.run is not None and not isinstance(self.run, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.run = str(self.run)
        if self.url is not None and not isinstance(self.url, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.url = str(self.url)
        if self.framework is not None and not isinstance(
            self.framework, Framework
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.framework = None

    def __repr__(self) -> str:
        return f"AdditionalInfo(url='{self.url}', run_id='{self.run}', framework='{self.framework}', extra_info={self.extra_info})"

    def _format_extra_info(self) -> list[dict[str, str]] | None:
        if self.extra_info is None:
            return None
        elif isinstance(self.extra_info, ExtraInfo):
            return [self.extra_info.asdict()]
        elif isinstance(self.extra_info, list):  # pyright: ignore[reportUnnecessaryIsInstance]
            return [info.asdict() for info in self.extra_info]
        else:
            raise ValueError("Invalid type for extra_info. Should be ExtraInfo, List[ExtraInfo], or None.")

    def asdict(self) -> dict[str, str | Framework | list[dict[str, str]] | None]:
        extra_info_dict = self._format_extra_info()
        return {
            "url": self.url,
            "run": self.run,
            "library": self.framework.value if self.framework is not None else None,
            "extraInfo": extra_info_dict,
        }


class Framework(Enum):
    """Enumeration of the different Experiment Tracker framework."""

    MLFLOW = "MLFLOW"
    WANDB = "WANDB"
    OTHER = "OTHER"


@dataclass
class ExtraInfo:
    """Represent a key value pair of extra information."""

    key: str
    value: str

    def __post_init__(self):
        if not isinstance(self.key, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.key = str(self.key)
        if not isinstance(self.value, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.value = str(self.value)

    def asdict(self) -> dict[str, str]:
        return {"key": self.key, "value": self.value}



from __future__ import annotations

from typing import Any

from vectice.api.json.json_type import TJSON
from vectice.models.resource.metadata.base import DatasetSourceType, DatasetSourceUsage


class DatasetResourceRepresentationOutput(TJSON):
    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def type(self) -> DatasetSourceType:
        return DatasetSourceType(self["type"])

    @property
    def usage(self) -> str:
        return DatasetSourceUsage(self["usage"]).value if self["usage"] else "GENERIC"

    @property
    def path(self) -> str:
        return str(self["uri"])

    @property
    def size(self) -> int | None:
        return int(self["size"]) if self["size"] else None

    @property
    def name(self) -> str:
        return self["filename"] if "filename" in self else self["tablename"]

    @property
    def columns_count(self) -> int | None:
        return int(self["columnsNumber"]) if self["columnsNumber"] else None

    @property
    def rows_count(self) -> int | None:
        return int(self["rowsNumber"]) if self["rowsNumber"] else None

    @property
    def info(self) -> dict[str, Any] | None:
        return {item["displayName"]: item["value"] for item in self["extraMetadata"]} if self["extraMetadata"] else None



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.service_types.vectice_types import VecticeType
from vectice.autolog.asset_services.technique_service import TechniqueService
from vectice.autolog.model_library import ModelLibrary
from vectice.models.model import Model
from vectice.utils.common_utils import get_asset_name

if TYPE_CHECKING:
    from statsmodels.base.wrapper import ResultsWrapper


class AutologStatsModelWrapperService(MetricService, TechniqueService):
    def __init__(
        self,
        key: str,
        asset: ResultsWrapper,
        data: dict,
        custom_metrics_data: set[str | None],
        phase_name: str,
        prefix: str | None = None,
    ):
        self._asset = asset
        self._key = key
        self._model_name = get_asset_name(self._key, phase_name, prefix)

        super().__init__(cell_data=data, custom_metrics=custom_metrics_data)

    def get_asset(self):
        _, params = self._get_statsmodel_info(self._asset)
        model_summary = self._asset.summary().as_text()
        model = Model(
            library=ModelLibrary.STATSMODEL.value,
            technique=self._get_model_technique(self._asset, ModelLibrary.STATSMODEL),
            metrics=self._get_statsmodels_metrics(self._asset),
            properties=params,
            name=self._model_name,
            predictor=self._asset,
        )
        return {
            "variable": self._key,
            "model": model,
            "summary": model_summary,
            "asset_type": VecticeType.MODEL,
        }

    def _get_statsmodel_info(self, model: ResultsWrapper) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        try:
            # model params
            model_params = {"model_type": model.model.__class__.__name__, "formula": model.model.formula}
            # statsmodels fitted/Resultswrapper
            params = {str(key): value for key, value in model.params.items() if value is not None}
            model_params.update(params)
            return "statsmodel", model_params
        except AttributeError:
            return None, None



from __future__ import annotations

import logging
import warnings

from vectice.api.http_error_handlers import VecticeException
from vectice.models.dataset import Dataset
from vectice.models.model import Model
from vectice.models.table import Table
from vectice.trace.code_trace_service import CodeTraceService
from vectice.trace.graph_patch_service import GraphPatchService
from vectice.trace.trace_client_service import TraceClientService
from vectice.utils.common_utils import get_asset_name

_logger = logging.getLogger(__name__)


class VecticeTrace(CodeTraceService):
    """VecticeTrace is a context manager that integrates Vectice's tracing capabilities with Python's code tracing.

    It captures the execution of Python code, including variables and function calls, and logs Vectice assets
    such as datasets, models, and tables to the current iteration in Vectice.

    ```python
    with VecticeTrace(
            phase = 'PHA-XXX'           # Paste your Phase Id)
        as trace:
        # Your code here
        dataset = Dataset(name="example_dataset")
    ```

    Environmental variables can be set for the trace configuration in a .env file.
    ```
    VECTICE_API_TOKEN='your-api-key'
    VECTICE_HOST='your-host-url'
    ```

    The environmental variables can also be set in the CLI.
    ```bash
    export VECTICE_API_TOKEN='your-api-key'
    export VECTICE_HOST='your-host-url'
    ```

    The environmental variables can also be set in directly.
    ```python
    import os
    os.environ['VECTICE_API_TOKEN'] = token
    os.environ['VECTICE_HOST'] = host
    ```

    Parameters:
        phase: The ID of the phase in which you wish to autolog your work as an iteration.
        prefix: A Prefix which will be applied to logged models and datasets variable name. e.g `my-prefix` will be `my-prefix-variable`,
            an empty string will be `variable` and no prefix will be `PHA-XXX-variable`.
        create_new_iteration: If set to False, logging of assets will happen in the last updated iteration. Otherwise, it will create a new iteration for logging the cell's assets.

    """

    def __init__(
        self,
        phase: str | None = None,
        prefix: str | None = None,
        create_new_iteration: bool = False,
    ):
        if self._is_ipython_environment():
            raise VecticeException(
                "Ipython environments are currently not supported by Vectice Trace. Please use Autolog instead."
                "For detailed information about autolog, supported libraries and environments please consult the documentation: https://api-docs.vectice.com/reference/vectice/autolog/"
            )
        CodeTraceService.__init__(self)
        self._client = TraceClientService(phase, create_new_iteration)
        self._graph_patch_service = GraphPatchService()
        self._failed_assets = []
        self._prefix = prefix

    def __enter__(self):
        """Enhanced context manager that coordinates tracing."""
        self._graph_patch_service.patch_all()

        self._client.is_vectice_client_valid()
        # Start tracing first
        CodeTraceService.__enter__(self)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # pyright: ignore[reportMissingParameterType]
        """Enhanced context manager that handles logging of assets and tracing cleanup."""
        CodeTraceService.__exit__(self, exc_type, exc_val, exc_tb)
        try:
            self._log_all_assets()
        except Exception as e:
            # Log API errors but don't prevent trace cleanup
            warnings.warn(f"Error during Vectice trace logging: {e}", RuntimeWarning, 1)
        finally:
            # Ensure cleanup of the graph patch service
            self._graph_patch_service.unpatch_all()
            return False

    #### Configuration ####

    def config(
        self, phase: str | None = None, prefix: str | None = None, create_new_iteration: bool | None = None
    ) -> VecticeTrace:
        """Configures the trace functionality within Vectice.

        The `config` method allows you to configure your Vectice trace by specifying the phase in which you want to trace your work.

        ```python
        # Configure trace functionality
        import vectice

        trace = vectice.trace()
        trace = trace.config(
            phase = 'PHA-XXX',           # Paste your Phase Id
            create_new_iteration = True
        )
        ```
        Parameters:
                phase: The ID of the phase in which you wish to autolog your work as an iteration.
                prefix: A Prefix which will be applied to logged models and datasets variable name. e.g `my-prefix` will be `my-prefix-variable`,
                    an empty string will be `variable` and no prefix will be `PHA-XXX-variable`.
                create_new_iteration: If set to False, logging of assets will happen in the last updated iteration. Otherwise, it will create a new iteration for logging the cell's assets.
        """
        self._client.config(phase, create_new_iteration)
        if prefix:
            self._prefix = prefix
        return self

    #### Utility Methods ####

    def _get_vectice_objects(self):
        """Get the Vectice objects that are currently being traced."""
        supported_types = (Dataset, Model, Table)
        vectice_assets = [asset for asset in self.captured_variables.values() if isinstance(asset, supported_types)]
        return vectice_assets

    def _is_ipython_environment(self) -> bool:
        """Prevent the Vectice Trace from being used in Jupyter Notebook environments."""
        try:
            from IPython.core.getipython import get_ipython

            ipython = get_ipython()
            return True if ipython else False
        except ImportError:
            pass
        return False

    #### Logging Methods ####

    def _log_all_assets(self) -> None:
        """Log all Vectice assets to the current iteration."""
        if not self._client.iteration_service:
            _logger.warning("No active iteration found.")
            return

        assets_to_log = self._get_vectice_objects()

        for asset in assets_to_log:
            self._log_vectice_asset(asset)

        patched_graphs = self._graph_patch_service.graphs
        if patched_graphs:
            self._log_graphs(patched_graphs)

        self._client.save_autolog_assets(cells=self.code, prefix=self._prefix, is_trace=True)

        if len(self._failed_assets):
            _logger.warning("The following assets failed to log:")
        for failed_asset in self._failed_assets:
            _logger.warning(f"{failed_asset['type']} {failed_asset['asset']!r}, reason: {failed_asset['reason']}")

    def _log_vectice_asset(self, asset_to_log: Dataset | Table | Model) -> None:
        """Log a Vectice asset to the current iteration."""
        try:
            asset_to_log.name = get_asset_name(str(asset_to_log.name), self._client.phase.name, self._prefix)
            object_name = asset_to_log.name

            if isinstance(asset_to_log, Dataset):
                self._client.iteration_service.log_dataset(asset_to_log)
                _logger.info(f"Dataset {object_name!r} logged in iteration {self._client.iteration.name!r}.")
            elif isinstance(asset_to_log, Table):
                self._client.iteration_service.log_table(asset_to_log)
                _logger.info(f"Table {object_name!r} logged in iteration {self._client.iteration.index!r}.")
            else:
                if asset_to_log.predictor:
                    self._client.iteration._assign_model_predictor_metadata(  # pyright: ignore[reportPrivateUsage]
                        asset_to_log
                    )
                self._client.iteration_service.log_model(asset_to_log)
                _logger.info(f"Model {object_name!r} logged in iteration {self._client.iteration.index!r}.")
        except Exception as e:
            self._failed_assets.append({"reason": e, "asset": asset_to_log, "type": type(asset_to_log).__name__})

    def _log_graphs(self, graphs: list[str]) -> None:
        for graph in graphs:
            try:
                self._client.iteration_service.log_image_or_file(graph)
                _logger.info(f"Graph {graph!r} logged in iteration {self._client.iteration.name!r}.")
            except Exception as e:
                self._failed_assets.append({"reason": e, "asset": graph, "type": "Graph"})



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pandas import DataFrame


class Table:
    """Table wraps a dataframe into a table format, to enable the logging of custom metadata within Vectice.

    This class formats metadata into a tabular structure, enabling the ability to log tables for documentation purposes.
    Once logged, this structured metadata is accessible within the Vectice app.

    NOTE: **IMPORTANT INFORMATION**
        Vectice stores the metadata in its original form, ensuring it remains unaltered without any processing or transformation.

    """

    def __init__(
        self,
        dataframe: DataFrame,
        name: str | None = None,
    ):
        """Wrap a Table.

        A Vectice Table is a wrapped pandas dataframe in a table format, which can then be logged to a Vectice iteration. (Maximum of 100 rows and 20 columns).

        Parameters:
            dataframe: The pandas dataframe to be wrapped as table.
            name: The name of the table for future reference.

        """
        self._name = name or f"table {datetime.now()}"
        self._dataframe = dataframe

    @property
    def name(self) -> str:
        """The table's name.

        Returns:
            The table name.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """Set the table's name.

        Parameters:
            name: The name of the table.
        """
        self._name = name

    @property
    def dataframe(self) -> DataFrame:
        """The table's data.

        Returns:
            dataframe: The pandas dataframe to be displayed as table.
        """
        return self._dataframe

    @dataframe.setter
    def dataframe(self, dataframe: DataFrame):
        """Set the table's pandas dataframe.

        Parameters:
            dataframe: The pandas dataframe to be displayed as table.
        """
        self._dataframe = dataframe
